import inspect  # Used for storing the input
from warnings import warn

import numpy as np
import pandas as pd

from ttim.aquifer import Aquifer, SimpleAquifer
from ttim.aquifer_parameters import param_3d, param_maq
from ttim.invlapnumba import compute_laplace_parameters_numba, invlap, invlapcomp
from ttim.plots import PlotTtim


class TimModel:
    def __init__(
        self,
        kaq=[1, 1],
        z=[3, 2, 1],
        Haq=[1, 1],
        Hll=[0],
        c=[1e100, 100],
        Saq=[1e-4, 1e-4],
        Sll=[0],
        poraq=0.3,
        porll=0.3,
        ltype=["a", "a"],
        topboundary="conf",
        phreatictop=False,
        tmin=1,
        tmax=10,
        tstart=0,
        M=10,
        kzoverkh=None,
        model3d=False,
        timmlmodel=None,
    ):
        self.elementlist = []
        self.elementdict = {}
        self.vbclist = []  # variable boundary condition 'v' elements
        self.zbclist = []  # zero and constant boundary condition 'z' elements
        self.gbclist = []  # given boundary condition 'g' elements
        # note: given bc elements don't have any unknowns
        self.tmin = tmin
        self.tmax = tmax
        self.tstart = tstart
        self.M = M
        self.aq = Aquifer(
            self,
            kaq,
            z,
            Haq,
            Hll,
            c,
            Saq,
            Sll,
            poraq,
            porll,
            ltype,
            topboundary,
            phreatictop,
            kzoverkh,
            model3d,
        )
        self.compute_laplace_parameters()
        self.name = "TimModel"
        self.modelname = "ml"  # Used for writing out input
        self.timmlmodel = timmlmodel
        if self.timmlmodel is not None:
            self.timmlmodel.solve()

        self.plots = PlotTtim(self)
        self.plot = self.plots.topview

        # NOTE: reinstate later, after deprecation below is removed?
        # self.xsection = self.plots.xsection

    def xsection(self, *args, **kwargs):
        raise DeprecationWarning(
            "This method is deprecated. Use `ml.plots.head_along_line()` instead."
        )

    def contour(self, *args, **kwargs):
        warn(
            category=DeprecationWarning,
            message="This method is deprecated. Use `ml.plots.contour()` instead.",
            stacklevel=1,
        )
        self.plots.contour(*args, **kwargs)

    def __repr__(self):
        return (
            f"{self.name}: {self.aq.naq} aquifer(s), tmin={self.tmin}, tmax={self.tmax}"
        )

    def initialize(self):
        # remove inhom elements (they are added again by inhom initialize)
        self.gbclist = [e for e in self.gbclist if not e.inhomelement]
        self.vbclist = [e for e in self.vbclist if not e.inhomelement]
        self.zbclist = [e for e in self.zbclist if not e.inhomelement]
        self.aq.initialize()
        self.gvbclist = self.gbclist + self.vbclist
        self.vzbclist = self.vbclist + self.zbclist
        # Given elements are first in list
        self.elementlist = self.gbclist + self.vbclist + self.zbclist
        self.ngbc = len(self.gbclist)
        self.nvbc = len(self.vbclist)
        self.nzbc = len(self.zbclist)
        self.ngvbc = self.ngbc + self.nvbc
        for e in self.elementlist:
            e.initialize()
        # lists used for inverse transform
        enumber = []
        etstart = []
        ebc = []
        for k in range(self.ngvbc):
            e = self.gvbclist[k]
            enumber.extend(len(e.tstart) * [k])
            etstart.extend(list(e.tstart))
            ebc.extend(list(e.bc))
        self.enumber = np.array(enumber)
        self.etstart = np.array(etstart)
        self.ebc = np.array(ebc)

    def addelement(self, e):
        if e.label is not None:
            self.elementdict[e.label] = e
        if e.type == "g":
            self.gbclist.append(e)
        elif e.type == "v":
            self.vbclist.append(e)
        elif e.type == "z":
            self.zbclist.append(e)

    def removeelement(self, e):
        if e.label is not None:
            self.elementdict.pop(e.label)
        if e.type == "g":
            self.gbclist.remove(e)
        elif e.type == "v":
            self.vbclist.remove(e)
        elif e.type == "z":
            self.zbclist.remove(e)

    def compute_laplace_parameters(self):
        """Compute the parameters for the Laplace transform inversion.

        Attributes
        ----------
        nint: Number of time intervals
        npint: Number of p values per interval
        npval: Total number of p values (nint * npint)
        p[npval]: Array with p values
        """
        itmin = np.floor(np.log10(self.tmin))
        itmax = np.ceil(np.log10(self.tmax))
        self.tintervals = 10.0 ** np.arange(itmin, itmax + 1)
        # lower and upper limit are adjusted to prevent any problems from t
        # exactly at the beginning and end of the interval
        # also, you cannot count on t >= 10 ** log10(t) for all possible t
        self.tintervals[0] = self.tintervals[0] * (1 - 1e-8)
        self.tintervals[-1] = self.tintervals[-1] * (1 + 1e-8)
        self.nint = len(self.tintervals) - 1  # number of p-intervals
        self.npint = 2 * self.M + 1  # number of p values in an interval
        self.npval = self.nint * self.npint
        # numba
        self.p = np.zeros((self.nint, self.npint), dtype=np.complex128)
        for i in range(self.nint):
            self.p[i] = compute_laplace_parameters_numba(self.tintervals[i + 1], self.M)
        # TODO: make self.p a 2D array
        self.p = np.ravel(self.p)
        # self.aq.initialize()

    def potential(self, x, y, t, layers=None, aq=None, derivative=0, returnphi=0):
        """Returns pot[naq, ntimes] if layers=None, otherwise pot[len(layers), ntimes].

        t must be ordered.
        """
        if aq is None:
            aq = self.aq.find_aquifer_data(x, y)
        if layers is None:
            layers = range(aq.naq)
        nlayers = len(layers)
        time = np.atleast_1d(t) - self.tstart  # used to be ).copy()
        pot = np.zeros((self.ngvbc, aq.naq, self.npval), dtype=complex)
        for i in range(self.ngbc):
            pot[i, :] += self.gbclist[i].unitpotential(x, y, aq)
        for e in self.vzbclist:
            pot += e.potential(x, y, aq)
        if layers is None:
            pot = np.sum(pot[:, np.newaxis, :, :] * aq.eigvec, 2)
        else:
            pot = np.sum(pot[:, np.newaxis, :, :] * aq.eigvec[layers, :], 2)
        if derivative > 0:
            pot *= self.p**derivative
        if returnphi:
            return pot
        rv = invlapcomp(
            time,
            pot,
            self.npint,
            self.M,
            self.tintervals,
            self.enumber,
            self.etstart,
            self.ebc,
            nlayers,
        )
        return rv

    def potentialone(self, x, y, time, layers=None, aq=None, derivative=0, returnphi=0):
        """Returns pot[naq] if layers=None, otherwise pot[len(layers)].

        time is one value.
        """
        if aq is None:
            aq = self.aq.find_aquifer_data(x, y)
        if layers is None:
            layers = range(aq.naq)
        nlayers = len(layers)
        time = np.atleast_1d(time) - self.tstart  # used to be ).copy()
        jtime = np.searchsorted(self.tintervals, time)[0] - 1
        assert 0 <= jtime <= len(self.tintervals), "time not in tintervals"
        pot = np.zeros((self.ngvbc, aq.naq, self.npint), dtype=complex)
        for i in range(self.ngbc):
            pot[i, :] += self.gbclist[i].unitpotentialone(x, y, jtime, aq)
        for e in self.vzbclist:
            pot += e.potential(x, y, aq)
        if layers is None:
            pot = np.sum(pot[:, np.newaxis, :, :] * aq.eigvec2[:, :, jtime], 2)
        else:
            pot = np.sum(pot[:, np.newaxis, :, :] * aq.eigvec2[layers, :, jtime], 2)
        if derivative > 0:
            pot *= self.p**derivative
        if returnphi:
            return pot
        rv = invlapcomp(
            time,
            pot[:, :, :],
            self.npint,
            self.M,
            self.tintervals[jtime : jtime + 2],
            self.enumber,
            self.etstart,
            self.ebc,
            nlayers,
        )
        return rv

    def disvec(self, x, y, t, layers=None, aq=None, derivative=0):
        """Compute discharge vectgor.

        Returns qx[naq, ntimes], qy[naq, ntimes] if layers=None, otherwise
        qx[len(layers,Ntimes)],qy[len(layers, ntimes)].

        t must be ordered.
        """
        if aq is None:
            aq = self.aq.find_aquifer_data(x, y)
        if layers is None:
            layers = range(aq.naq)
        nlayers = len(layers)
        time = np.atleast_1d(t) - self.tstart
        disx = np.zeros((self.ngvbc, aq.naq, self.npval), dtype=complex)
        disy = np.zeros((self.ngvbc, aq.naq, self.npval), dtype=complex)
        for i in range(self.ngbc):
            qx, qy = self.gbclist[i].unitdisvec(x, y, aq)
            disx[i, :] += qx
            disy[i, :] += qy
        for e in self.vzbclist:
            qx, qy = e.disvec(x, y, aq)
            disx += qx
            disy += qy
        if layers is None:
            disx = np.sum(disx[:, np.newaxis, :, :] * aq.eigvec, 2)
            disy = np.sum(disy[:, np.newaxis, :, :] * aq.eigvec, 2)
        else:
            disx = np.sum(disx[:, np.newaxis, :, :] * aq.eigvec[layers, :], 2)
            disy = np.sum(disy[:, np.newaxis, :, :] * aq.eigvec[layers, :], 2)
        if derivative > 0:
            disx *= self.p**derivative
            disy *= self.p**derivative
        rvx = invlapcomp(
            time,
            disx,
            self.npint,
            self.M,
            self.tintervals,
            self.enumber,
            self.etstart,
            self.ebc,
            nlayers,
        )
        rvy = invlapcomp(
            time,
            disy,
            self.npint,
            self.M,
            self.tintervals,
            self.enumber,
            self.etstart,
            self.ebc,
            nlayers,
        )
        return rvx, rvy

    def head(self, x, y, t, layers=None, aq=None, derivative=0, neglect_steady=False):
        """Head at x, y, t where t can be multiple times.

        Parameters
        ----------
        x : float
        y : float
        t : list or array
            times for which grid is returned
        layers : integer, list or array, optional
            layers for which grid is returned
            if None: all layers are returned

        Returns
        -------
        h : array size `nlayers, ntimes`
        """
        if aq is None:
            aq = self.aq.find_aquifer_data(x, y)
        if layers is None:
            layers = range(aq.naq)
        else:
            layers = np.atleast_1d(layers)  # corrected for base zero
        pot = self.potential(x, y, t, layers, aq, derivative)
        h = aq.potential_to_head(pot, layers)
        if self.timmlmodel is not None:
            if not neglect_steady:
                htimml = self.timmlmodel.head(x, y, layers=layers)
                h += htimml[:, np.newaxis]
        return h

    def velocompold(self, x, y, z, t, aq=None, layer_ltype=[0, 0]):
        # implemented for one layer
        if aq is None:
            aq = self.aq.find_aquifer_data(x, y)
        assert (z <= aq.z[0]) & (z >= aq.z[-1]), "z value not inside aquifer"
        if layer_ltype is None:
            layer, ltype, dummy = aq.findlayer(z)
        else:
            layer, ltype = layer_ltype
        qx, qy = self.disvec(x, y, t, aq=aq)
        layer = layer_ltype[0]
        vx = qx[layer] / (aq.Haq[layer] * aq.poraq[layer])
        vy = qy[layer] / (aq.Haq[layer] * aq.poraq[layer])
        vz = np.zeros_like(vx)
        return vx, vy, vz

    def velocomp(self, x, y, z, t, aq=None, layer_ltype=None):
        # compute velocity for one point x, y, z, t
        if aq is None:
            aq = self.aq.find_aquifer_data(x, y)
        assert (z <= aq.z[0]) & (z >= aq.z[-1]), "z value not inside aquifer"
        if layer_ltype is None:
            layer, ltype, dummy = aq.findlayer(z)
        else:
            layer, ltype = layer_ltype
        if ltype == "l":  # inside leaky layer
            vx = 0.0
            vy = 0.0
            if layer == 0:
                h = self.head(x, y, t, layers=layer, aq=aq, neglect_steady=True)
                qz = (h[0, 0] - 0.0) / aq.c[0]
            else:
                h = self.head(
                    x, y, t, layers=[layer - 1, layer], aq=aq, neglect_steady=True
                )
                qz = (h[1, 0] - h[0, 0]) / aq.c[
                    layer
                ]  # TODO: include storage in leaky layer
            vz = qz / aq.porll[layer]
        else:  # in aquifer layer
            h = self.head(x, y, t, layers=layer, aq=aq, neglect_steady=True)
            qx, qy = self.disvec(x, y, t, aq=aq)
            vx = qx[layer, 0] / (
                aq.Haq[layer] * (aq.poraq[layer] + aq.Saq[layer] * h[0, 0])
            )
            vy = qy[layer, 0] / (
                aq.Haq[layer] * (aq.poraq[layer] + aq.Saq[layer] * h[0, 0])
            )
            #
            h = np.zeros(3)  # head above layer, in layer, and below layer
            if layer > 0:
                if layer < aq.naq - 1:  # there is a layer above and below
                    h[:] = self.head(
                        x,
                        y,
                        t,
                        layers=[layer - 1, layer, layer + 1],
                        aq=aq,
                        neglect_steady=True,
                    )[:, 0]
                else:
                    h[:2] = self.head(
                        x, y, t, layers=[layer - 1, layer], aq=aq, neglect_steady=True
                    )[:, 0]
            else:  # layer = 0, so top layer
                if aq.naq == 1:  # only one layer
                    h[1] = self.head(x, y, t, layers=[layer], aq=aq, neglect_steady=True)[
                        :, 0
                    ]
                else:
                    h[1:] = self.head(
                        x, y, t, layers=[layer, layer + 1], aq=aq, neglect_steady=True
                    )[:, 0]
            # this works because c[0] = 1e100 for impermeable top
            qztop = (h[1] - h[0]) / self.aq.c[layer]
            # TODO: modify for infiltration in top aquifer
            # if layer == 0:
            #    qztop += self.qztop(x, y)
            if layer < aq.naq - 1:
                qzbot = (h[2] - h[1]) / self.aq.c[layer + 1]
            else:
                qzbot = 0.0
            vz = (
                qzbot + (z - aq.zaqbot[layer]) / aq.Haq[layer] * (qztop - qzbot)
            ) / aq.poraq[layer]
        velo = np.array([vx, vy, vz])

        if self.timmlmodel is not None:
            velotimml = self.timmlmodel.velocity(x, y, z)
            velo += velotimml

        return velo

    def velo_one(self, x, y, z, t, aq=None, layer_ltype=[0, 0]):
        # implemented for one layer and one time
        vx, vy, vz = self.velocomp(x, y, z, t, aq, layer_ltype)
        return np.array([vx[0], vy[0], vz[0]])

    def headinside(self, elabel, t):
        return self.elementdict[elabel].headinside(t - self.tstart)

    def strength(self, elabel, t):
        return self.elementdict[elabel].strength(t - self.tstart)

    def headalongline(self, x, y, t, layers=None):
        """Head along line or curve.

        Parameters
        ----------
        x : 1D array or list
            x values of line
        y : 1D array or list
            y values of line
        t : float or 1D array or list
            times for which grid is returned
        layers : integer, list or array, optional
            layers for which grid is returned

        Returns
        -------
        h : array size `nlayers, ntimes, nx`
        """
        xg = np.atleast_1d(x)
        yg = np.atleast_1d(y)
        if layers is None:
            Nlayers = self.aq.find_aquifer_data(xg[0], yg[0]).naq
        else:
            Nlayers = len(np.atleast_1d(layers))
        nx = len(xg)
        if len(yg) == 1:
            yg = yg * np.ones(nx)
        t = np.atleast_1d(t)
        h = np.zeros((Nlayers, len(t), nx))
        for i in range(nx):
            h[:, :, i] = self.head(xg[i], yg[i], t, layers)
        return h

    def disvecalongline(self, x, y, t, layers=None):
        """Discharge vector along line or curve.

        Parameters
        ----------
        x : 1D array or list
            x values of line
        y : 1D array or list
            y values of line
        t : float or 1D array or list
            times for which grid is returned
        layers : integer, list or array, optional
            layers for which grid is returned

        Returns
        -------
        q : array size `nlayers, ntimes, nx`
        """
        xg = np.atleast_1d(x)
        yg = np.atleast_1d(y)
        if layers is None:
            nlay = self.aq.find_aquifer_data(xg[0], yg[0]).naq
        else:
            nlay = len(np.atleast_1d(layers))
        nx = len(xg)
        if len(yg) == 1:
            yg = yg * np.ones(nx)
        t = np.atleast_1d(t)
        qx = np.zeros((nlay, len(t), nx))
        qy = np.zeros((nlay, len(t), nx))
        for i in range(nx):
            qx[:, :, i], qy[:, :, i] = self.disvec(xg[i], yg[i], t, layers)
        return qx, qy

    def headgrid(self, xg, yg, t, layers=None, printrow=False):
        """Grid of heads.

        Parameters
        ----------
        xg : array
            x values of grid
        yg : array
            y values of grid
        t : list or array
            times for which grid is returned
        layers : integer, list or array, optional
            layers for which grid is returned
        printrow : boolean, optional
            prints dot to screen for each row of grid if set to `True`

        Returns
        -------
        h : array size `nlayers, ntimes, ny, nx`

        See Also
        --------
        :func:`~ttim.model.Model.headgrid2`
        """
        nx = len(xg)
        ny = len(yg)
        if layers is None:
            nlayers = self.aq.find_aquifer_data(xg[0], yg[0]).naq
        else:
            nlayers = len(np.atleast_1d(layers))
        t = np.atleast_1d(t)
        h = np.empty((nlayers, len(t), ny, nx))
        for j in range(ny):
            if printrow:
                print(".", end="", flush=True)
            for i in range(nx):
                h[:, :, j, i] = self.head(xg[i], yg[j], t, layers)
        return h

    def headgrid2(self, x1, x2, nx, y1, y2, ny, t, layers=None, printrow=False):
        """Grid of heads.

        Parameters
        ----------
        xg : array
            x values are generated as linspace(x1, x2, nx)
        yg : array
            y values are generated as linspace(y1, y2, ny)
        t : list or array
            times for which grid is returned
        layers : integer, list or array, optional
            layers for which grid is returned
        printrow : boolean, optional
            prints dot to screen for each row of grid if set to `True`

        Returns
        -------
        h : array size `nlayers, ntimes, ny, nx`

        See Also
        --------
        :func:`~ttim.model.Model.headgrid`
        """
        xg = np.linspace(x1, x2, nx)
        yg = np.linspace(y1, y2, ny)
        return self.headgrid(xg, yg, t, layers, printrow)

    def inverseLapTran(self, pot, t):
        """Returns array of potentials of len(t) t must be ordered and tmin<=t<=tmax."""
        t = np.atleast_1d(t)
        rv = np.zeros(len(t))
        it = 0
        if t[-1] >= self.tmin:  # Otherwise all zero
            if t[0] < self.tmin:
                it = np.argmax(t >= self.tmin)
            for n in range(self.nint):
                if n == self.nint - 1:
                    tp = t[(t >= self.tintervals[n]) & (t <= self.tintervals[n + 1])]
                else:
                    tp = t[(t >= self.tintervals[n]) & (t < self.tintervals[n + 1])]
                nt = len(tp)
                if nt > 0:  # if all values zero, don't do inverse transform
                    # Not needed anymore: if np.abs(pot[n*self.npint]) > 1e-20:
                    # If there is a zero item, zero should be returned;
                    # funky enough this can be done with a
                    # straight equal comparison
                    if not np.any(pot[n * self.npint : (n + 1) * self.npint] == 0.0):
                        rv[it : it + nt] = invlap(
                            tp,
                            self.tintervals[n + 1],
                            pot[n * self.npint : (n + 1) * self.npint],
                            self.M,
                        )
                    it = it + nt
        return rv

    def solve(self, printmat=0, sendback=0, silent=False):
        """Compute solution."""
        # Initialize elements
        self.initialize()
        # Compute number of equations
        self.neq = np.sum([e.nunknowns for e in self.elementlist])
        if silent is False:
            print("self.neq ", self.neq)
        if self.neq == 0:
            if silent is False:
                print("No unknowns. Solution complete")
            return
        mat = np.empty((self.neq, self.neq, self.npval), dtype=complex)
        rhs = np.empty((self.neq, self.ngvbc, self.npval), dtype=complex)
        ieq = 0
        for e in self.elementlist:
            if e.nunknowns > 0:
                (
                    mat[ieq : ieq + e.nunknowns, :, :],
                    rhs[ieq : ieq + e.nunknowns, :, :],
                ) = e.equation()
                ieq += e.nunknowns
        if printmat:
            return mat, rhs
        for i in range(self.npval):
            sol = np.linalg.solve(mat[:, :, i], rhs[:, :, i])
            icount = 0
            for e in self.elementlist:
                for j in range(e.nunknowns):
                    e.parameters[:, j, i] = sol[icount, :]
                    icount += 1
                e.run_after_solve()
        if silent is False:
            print("solution complete")
        elif (silent == "dot") or (silent == "."):
            print(".", end="", flush=True)
        if sendback:
            return sol
        return

    def storeinput(self, frame):
        self.inputargs, _, _, self.inputvalues = inspect.getargvalues(frame)

    def write(self):
        rv = self.modelname + " = " + self.name + "(\n"
        for key in self.inputargs[1:]:  # The first argument (self) is ignored
            if isinstance(self.inputvalues[key], np.ndarray):
                rv += (
                    key
                    + " = "
                    + np.array2string(self.inputvalues[key], separator=",")
                    + ",\n"
                )
            elif isinstance(self.inputvalues[key], str):
                rv += key + " = '" + self.inputvalues[key] + "',\n"
            else:
                rv += key + " = " + str(self.inputvalues[key]) + ",\n"
        rv += ")\n"
        return rv

    def writemodel(self, fname):
        self.initialize()  # So that model can be written without solving first
        f = open(fname, "w")
        f.write("from ttim import *\n")
        f.write(self.write())
        for e in self.elementlist:
            f.write(e.write())
        f.close()

    def aquifer_summary(self):
        """Return DataFrame with summary of aquifer(s) parameters in model.

        Returns
        -------
        pandas.DataFrame
            dataframe with summary of aquifer(s) parameters
        """
        aqs = {}
        if not isinstance(self.aq, SimpleAquifer):
            aqs["background"] = self.aq.summary()
        for name, iaq in self.aq.inhomdict.items():
            aqs[name] = iaq.summary()
        return pd.concat(aqs, axis=0)


class ModelMaq(TimModel):
    """Create model specifying a multi-aquifer sequence of aquifer-leakylayer-etc.

    Parameters
    ----------
    kaq : float, array or list
        hydraulic conductivity of each aquifer from the top down
        if float, hydraulic conductivity is the same in all aquifers
    z : array or list
        elevation tops and bottoms of the aquifers from the top down
        leaky layers may have zero thickness
        if top='conf': length is 2 * number of aquifers
        if top='semi': length is 2 * number of aquifers + 1 as top
        of leaky layer on top of systems needs to be specified
    c : float, array or list
        resistance of leaky layers from the top down
        if float, resistance is the same for all leaky layers
        if top='conf': length is number of aquifers - 1
        if top='semi': length is number of aquifers
    Saq : float, array or list
        specific storage of all aquifers
        if float, sepcific storage is same in all aquifers
        if phreatictop is True and topboundary is 'conf', Saq of top
        aquifer is phreatic storage coefficient (and not multiplied
        with the layer thickness)
    Sll : float, array or list
        specific storage of all leaky layers
        if float, sepcific storage is same in all leaky layers
        if phreatictop is True and topboundary is 'semi', Sll of top
        leaky layer is phreatic storage coefficient (and not multiplied
        with the layer thickness)
    topboundary : string, 'conf' or 'semi' (default is 'conf')
        indicating whether the top is confined ('conf') or
        semi-confined ('semi')
    phreatictop : boolean
        the storage coefficient of the top model layer is treated as
        phreatic storage (and not multiplied with the aquifer thickness)
    tmin : scalar
        the minimum time for which heads can be computed after any change
        in boundary condition.
    tmax : scalar
        the maximum time for which heads can be computed
    tstart : scalar
        time at start of simulation (default 0)
    M : integer
        the number of terms to be used in the numerical inversion algorithm.
        10 is usually sufficient. If drawdown curves appear to oscillate,
        more terms may be needed, but this seldom happens.
    timmlmodel : optional instance of a solved TimML model
        a timml model may be included to add steady-state flow
    """

    def __init__(
        self,
        kaq=[1],
        z=[1, 0],
        c=[],
        Saq=[0.001],
        Sll=[0],
        poraq=[0.3],
        porll=[0.3],
        topboundary="conf",
        phreatictop=False,
        tmin=1,
        tmax=10,
        tstart=0,
        M=10,
        timmlmodel=None,
    ):
        self.storeinput(inspect.currentframe())
        kaq, Haq, Hll, c, Saq, Sll, poraq, porll, ltype = param_maq(
            kaq, z, c, Saq, Sll, poraq, porll, topboundary, phreatictop
        )
        super().__init__(
            kaq,
            z,
            Haq,
            Hll,
            c,
            Saq,
            Sll,
            poraq,
            porll,
            ltype,
            topboundary,
            phreatictop,
            tmin,
            tmax,
            tstart,
            M,
            timmlmodel=timmlmodel,
        )
        self.name = "ModelMaq"


class Model3D(TimModel):
    """Create a multi-layer model object consisting of many aquifer layers.

    The
    resistance between the layers is computed from the vertical hydraulic conductivity
    of the layers.

    Parameters
    ----------
    kaq : float, array or list
        hydraulic conductivity of each layer from the top down
        if float, hydraulic conductivity is the same in all aquifers
    z : array or list
        elevation of top of system followed by bottoms of all layers
        from the top down
        bottom of layer is automatically equal to top of layer below it
        if topboundary='conf': length is number of layers + 1
        if topboundary='semi': length is number of layers + 2 as top
        of leaky layer on top of systems needs to be specified
    Saq : float, array or list
        specific storage of all aquifers layers
        if float, sepcific storage is same in all aquifers layers
        if phreatictop is True and topboundary is 'conf', Saq of top
        aquifer is phreatic storage coefficient (and not multiplied
        with the layer thickness)
    kzoverkh : float
        vertical anisotropy ratio vertical k divided by horizontal k
        if float, value is the same for all layers
        length is number of layers
    topboundary : string, 'conf' or 'semi' (default is 'conf')
        indicating whether the top is confined ('conf') or
        semi-confined ('semi').
        currently only implemented for 'conf'
    topres : float
        resistance of top semi-confining layer, only read if topboundary='semi'
    topthick: float
        thickness of top semi-confining layer, only read if topboundary='semi'
    phreatictop : boolean
        the storage coefficient of the top aquifer layer is treated as
        phreatic storage (and not multiplied with the aquifer thickness)
    tmin : scalar
        the minimum time for which heads can be computed after any change
        in boundary condition.
    tmax : scalar
        the maximum time for which heads can be computed.
    tstart : scalar
        time at start of simulation (default 0)
    M : integer (default 10)
        the number of terms to be used in the numerical inversion algorithm.
        10 is usually sufficient. If drawdown curves appear to oscillate,
        more terms may be needed, but this seldom happens.
    timmlmodel : optional instance of a solved TimML model
        a timml model may be included to add steady-state flow
    """

    def __init__(
        self,
        kaq=1,
        z=[4, 3, 2, 1],
        Saq=0.001,
        kzoverkh=0.1,
        poraq=0.3,
        topboundary="conf",
        phreatictop=False,
        topres=0,
        topthick=0,
        topSll=0,
        toppor=0.3,
        tmin=1,
        tmax=10,
        tstart=0,
        M=10,
        timmlmodel=None,
    ):
        """Z must have the length of the number of layers + 1."""
        self.storeinput(inspect.currentframe())
        kaq, Haq, Hll, c, Saq, Sll, poraq, porll, ltype, z = param_3d(
            kaq,
            z,
            Saq,
            kzoverkh,
            poraq,
            phreatictop,
            topboundary,
            topres,
            topthick,
            topSll,
            toppor,
        )
        super().__init__(
            kaq,
            z,
            Haq,
            Hll,
            c,
            Saq,
            Sll,
            poraq,
            porll,
            ltype,
            topboundary,
            phreatictop,
            tmin,
            tmax,
            tstart,
            M,
            kzoverkh,
            model3d=True,
            timmlmodel=timmlmodel,
        )
        self.name = "Model3D"


class ModelXsection(TimModel):
    """Model class for cross-section models.

    Parameters
    ----------
    naq : integer
        number of aquifers
    tmin : float
        the minimum time for which heads can be computed after any change
        in boundary condition.
    tmax : float
        the maximum time for which heads can be computed.
    tstart : float, optional
        time at start of simulation (default 0)
    M : integer, optional
        the number of terms to be used in the numerical inversion algorithm.
        10 is usually sufficient.
    timmlmodel : timml.Model
        a timml model may be included to add a steady-state flow result to
        the computed solution.
    """

    def __init__(
        self,
        naq=1,
        tmin=1,
        tmax=10,
        tstart=0,
        M=10,
        timmlmodel=None,
    ):
        self.elementlist = []
        self.elementdict = {}
        self.vbclist = []  # variable boundary condition 'v' elements
        self.zbclist = []  # zero and constant boundary condition 'z' elements
        self.gbclist = []  # given boundary condition 'g' elements
        # note: given bc elements don't have any unknowns
        self.tmin = tmin
        self.tmax = tmax
        self.tstart = tstart
        self.M = M
        self.aq = SimpleAquifer(naq)
        self.compute_laplace_parameters()
        self.name = "TimModel"
        self.modelname = "ml"  # Used for writing out input
        self.timmlmodel = timmlmodel
        if self.timmlmodel is not None:
            self.timmlmodel.solve()

        self.plots = PlotTtim(self)
        self.plot = self.plots.topview
        self.name = "ModelXsection"

    def check_inhoms(self):
        """Check if number of aquifers in inhoms matches number of aquifers in model."""
        naqs = {}
        for inhom in self.aq.inhomdict.values():
            naqs[inhom.name] = inhom.naq
        check = np.array(list(naqs.values())) == self.aq.naq
        if not check.all():
            raise ValueError(f"Number of aquifers does not match {self.aq.naq}:\n{naqs}")
        # # shared boundary check
        # # NOTE: does not deal with nested inhoms
        # xcoords = np.concatenate(
        #     [(inhom.x1, inhom.x2) for inhom in self.aq.inhomdict.values()]
        # )
        # xcoords.sort()
        # if not np.all(np.diff(xcoords[1:-1])[::2] < 1e-10):
        #     raise ValueError("Not all inhomogeneities have shared boundaries.")

    def initialize(self):
        self.check_inhoms()
        super().initialize()



import numpy as np


def param_maq(
    kaq=[1],
    z=[1, 0],
    c=[],
    Saq=[0.001],
    Sll=[0],
    poraq=[0.3],
    porll=[0.3],
    topboundary="conf",
    phreatictop=False,
):
    # Computes the parameters for a TimModel from input for a maq model
    z = np.atleast_1d(z).astype(float)
    kaq = np.atleast_1d(kaq).astype(float)
    Saq = np.atleast_1d(Saq).astype(float)
    poraq = np.atleast_1d(poraq).astype(float)
    c = np.atleast_1d(c).astype(float)
    Sll = np.atleast_1d(Sll).astype(float)
    porll = np.atleast_1d(porll).astype(float)
    H = z[:-1] - z[1:]
    assert np.all(H >= 0), (
        "Error: Not all layer thicknesses are" + " non-negative" + str(H)
    )
    if topboundary[:3] == "con":
        assert len(z) % 2 == 0, (
            "Error: Length of z must be 2 * number of aquifers "
            "when topboundary is confined in MaqModel"
        )
        naq = int(len(z) / 2)
        if len(kaq) == 1:
            kaq = kaq * np.ones(naq)
        if len(Saq) == 1:
            Saq = Saq * np.ones(naq)
        if len(poraq) == 1:
            poraq = poraq * np.ones(naq)
        if len(c) == 1:
            c = c * np.ones(naq - 1)
        if len(Sll) == 1:
            Sll = Sll * np.ones(naq - 1)
        if len(porll) == 1:
            porll = porll * np.ones(naq - 1)
        assert len(kaq) == naq, "Error: Length of kaq needs to be " + str(naq)
        assert len(Saq) == naq, "Error: Length of Saq needs to be " + str(naq)
        assert len(poraq) == naq, "Error: Length of poraq needs to be " + str(naq)
        assert len(c) == naq - 1, "Error: Length of c needs to be " + str(naq - 1)
        assert len(Sll) == naq - 1, "Error: Length of Sll needs to be " + str(naq - 1)
        assert len(porll) == naq - 1, "Error: Length of porll needs to be " + str(naq - 1)
        Haq = H[::2]
        assert np.all(Haq > 0), (
            "Error: Some thicknesses of aquifer layers " + "are negative"
        )
        Hll = H[1::2]
        Hll = np.maximum(Hll, 1e-20)  # make sure none are negative
        assert np.all(Hll > 0), (
            "Error: Some thicknesses of leaky layers " + "are negative"
        )
        c = np.hstack((1e100, c))
        Sll = np.hstack((1e-20, Sll))
        Hll = np.hstack((1e-20, Hll))
        porll = np.hstack((1e-20, porll))
        # layertype
        nlayers = len(z) - 1
        ltype = np.array(nlayers * ["a"])
        ltype[1::2] = "l"
    else:  # leaky layers on top
        assert len(z) % 2 == 1, (
            "Error: Length of z must be 2 * number of aquifers + 1 "
            "when topboundary is leaky in MaqModel"
        )
        naq = int((len(z) - 1) / 2)
        if len(kaq) == 1:
            kaq = kaq * np.ones(naq)
        if len(Saq) == 1:
            Saq = Saq * np.ones(naq)
        if len(poraq) == 1:
            poraq = poraq * np.ones(naq)
        if len(c) == 1:
            c = c * np.ones(naq)
        if len(Sll) == 1:
            Sll = Sll * np.ones(naq)
        if len(porll) == 1:
            porll = porll * np.ones(naq)
        assert len(kaq) == naq, "Error: Length of kaq needs to be " + str(naq)
        assert len(Saq) == naq, "Error: Length of Saq needs to be " + str(naq)
        assert len(poraq) == naq, "Error: Length of poraq needs to be " + str(naq)
        assert len(c) == naq, "Error: Length of c needs to be " + str(naq)
        assert len(Sll) == naq, "Error: Length of Sll needs to be " + str(naq)
        assert len(porll) == naq, "Error: Length of porll needs to be " + str(naq)
        Haq = H[1::2]
        Hll = H[::2]
        # layertype
        nlayers = len(z) - 1
        ltype = np.array(nlayers * ["a"])
        ltype[0::2] = "l"
    return kaq, Haq, Hll, c, Saq, Sll, poraq, porll, ltype


def param_3d(
    kaq=[1],
    z=[1, 0],
    Saq=[0.001],
    kzoverkh=1,
    poraq=0.3,
    phreatictop=False,
    topboundary="conf",
    topres=0,
    topthick=0,
    topSll=0,
    toppor=0.3,
):
    # Computes the parameters for a TimModel from input for a 3D model
    kaq = np.atleast_1d(kaq).astype(float)
    z = np.atleast_1d(z).astype(float)
    naq = len(z) - 1
    if len(kaq) == 1:
        kaq = kaq * np.ones(naq)
    Saq = np.atleast_1d(Saq).astype(float)
    if len(Saq) == 1:
        Saq = Saq * np.ones(naq)
    kzoverkh = np.atleast_1d(kzoverkh).astype(float)
    if len(kzoverkh) == 1:
        kzoverkh = kzoverkh * np.ones(naq)
    poraq = np.atleast_1d(poraq).astype(float)
    if len(poraq) == 1:
        poraq = poraq * np.ones(naq)
    Haq = z[:-1] - z[1:]
    c = 0.5 * Haq[:-1] / (kzoverkh[:-1] * kaq[:-1]) + 0.5 * Haq[1:] / (
        kzoverkh[1:] * kaq[1:]
    )
    # Saq = Saq * H
    # if phreatictop:
    #    Saq[0] = Saq[0] / H[0]
    c = np.hstack((1e100, c))
    Hll = 1e-20 * np.ones(len(c))
    Sll = 1e-20 * np.ones(len(c))
    porll = np.zeros(len(c))
    nlayers = len(z) - 1
    ltype = np.array(nlayers * ["a"])
    if (topboundary[:3] == "sem") or (topboundary[:3] == "lea"):
        c[0] = np.max([1e-20, topres])
        Hll[0] = np.max([1e-20, topthick])
        Sll[0] = np.max([1e-20, topSll])
        porll[0] = toppor
        ltype = np.hstack(("l", ltype))
        z = np.hstack((z[0] + topthick, z))
    return kaq, Haq, Hll, c, Saq, Sll, poraq, porll, ltype, z



import numpy as np
from scipy.special import exp1

import ttim


def theis(r, t, T, S, Q):
    u = r**2 * S / (4 * T * t)
    h = -Q / (4 * np.pi * T) * exp1(u)
    return h


def theisQr(r, t, T, S, Q):
    u = r**2 * S / (4 * T * t)
    Qr = -Q / (2 * np.pi) * np.exp(-u) / r
    return Qr


T = 500
S = 1e-3
t = np.logspace(-4, 1, 10)
r = 30
Q = 788

h1 = theis(r, t, T, S, Q)

ml = ttim.ModelMaq(kaq=50, z=[10, 0], Saq=S / 10, tmin=1e-4, tmax=10)
w = ttim.Well(ml, 0, 0, rw=0.3, tsandQ=[(0, Q)])
ml.solve()
h2 = ml.head(r, 0, t)[0]

assert np.allclose(h1, h2, atol=1e-4), "h1 and h2 not all close Theis well 1"

# turn Theis off
t = np.logspace(-1, 1, 10)

h1 = theis(r, t, T, S, Q)
h1[t > 5] -= theis(r, t[t > 5] - 5, T, S, Q)

ml = ttim.ModelMaq(kaq=50, z=[10, 0], Saq=S / 10, tmin=1e-4, tmax=10)
w = ttim.Well(ml, 0, 0, rw=0.3, tsandQ=[(0, Q), (5, 0)])
ml.solve()
h2 = ml.head(r, 0, t)[0]

assert np.allclose(h1, h2, atol=1e-4), "h1 and h2 not all close Theis well 2"

# test nan values for Theis well 1
t = np.array([0.08, 0.09, 0.1, 1, 5, 9])
h1 = theis(r, t, T, S, Q)

tmin = 0.1
ml = ttim.ModelMaq(kaq=50, z=[10, 0], Saq=S / 10, tmin=tmin, tmax=10)
w = ttim.Well(ml, 0, 0, rw=0.3, tsandQ=[(0, Q)])
ml.solve()
h2 = ml.head(r, 0, t)[0]

a = np.isnan(h2)
b = t < tmin

assert np.all(a == b), "nans not in the right spot for tmin"
assert np.allclose(h1[~b], h2[~b], atol=1e-4)

# test nan values for Theis well 2

t = np.array([0.08, 0.09, 0.1, 1, 5, 5.03, 9])
h1 = theis(r, t, T, S, Q)
h1[t > 5] -= theis(r, t[t > 5] - 5, T, S, Q)

tmin = 0.1
ml = ttim.ModelMaq(kaq=50, z=[10, 0], Saq=S / 10, tmin=tmin, tmax=10)
w = ttim.Well(ml, 0, 0, rw=0.3, tsandQ=[(0, Q), (5, 0)])
ml.solve()
h2 = ml.head(r, 0, t)[0]

a = np.isnan(h2)
b = (t < tmin) | ((t > 5) & (t < 5.1))

assert np.all(a == b), "nans not in the right spot for tmin"
assert np.allclose(h1[~b], h2[~b], atol=1e-4)

### test for Qr ########################

Qr1 = theisQr(r, t, T, S, Q)

ml = ttim.ModelMaq(kaq=50, z=[10, 0], Saq=S / 10, tmin=1e-4, tmax=10)
w = ttim.Well(ml, 0, 0, rw=0.3, tsandQ=[(0, Q)])
ml.solve()
Qr2 = ml.disvec(r, 0, t)[0][0]

assert np.allclose(Qr1, Qr2, atol=1e-4), "Qr1 and Qr2 not all close Theis well 1"

# turn Theis off
t = np.logspace(-1, 1, 10)

Qr1 = theisQr(r, t, T, S, Q)
Qr1[t > 5] -= theisQr(r, t[t > 5] - 5, T, S, Q)

ml = ttim.ModelMaq(kaq=50, z=[10, 0], Saq=S / 10, tmin=1e-4, tmax=10)
w = ttim.Well(ml, 0, 0, rw=0.3, tsandQ=[(0, Q), (5, 0)])
ml.solve()
Qr2 = ml.disvec(r, 0, t)[0][0]

assert np.allclose(Qr1, Qr2, atol=1e-4), "Qr1 and Qr2 not all close Theis well 2"

# test nan values for Theis well 1
t = np.array([0.08, 0.09, 0.1, 1, 5, 9])
Qr1 = theisQr(r, t, T, S, Q)

tmin = 0.1
ml = ttim.ModelMaq(kaq=50, z=[10, 0], Saq=S / 10, tmin=tmin, tmax=10)
w = ttim.Well(ml, 0, 0, rw=0.3, tsandQ=[(0, Q)])
ml.solve()
Qr2 = ml.disvec(r, 0, t)[0][0]

a = np.isnan(Qr2)
b = t < tmin

assert np.all(a == b), "nans not in the right spot for tmin"
assert np.allclose(Qr1[~b], Qr2[~b], atol=1e-4)

# test nan values for Theis well 2

t = np.array([0.08, 0.09, 0.1, 1, 5, 5.03, 9])
Qr1 = theisQr(r, t, T, S, Q)
Qr1[t > 5] -= theisQr(r, t[t > 5] - 5, T, S, Q)

tmin = 0.1
ml = ttim.ModelMaq(kaq=50, z=[10, 0], Saq=S / 10, tmin=tmin, tmax=10)
w = ttim.Well(ml, 0, 0, rw=0.3, tsandQ=[(0, Q), (5, 0)])
ml.solve()
Qr2 = ml.disvec(r, 0, t)[0][0]

a = np.isnan(h2)
b = (t < tmin) | ((t > 5) & (t < 5.1))

assert np.all(a == b), "nans not in the right spot for tmin"
assert np.allclose(Qr1[~b], Qr2[~b], atol=1e-4)



import os

import nbformat
import pytest
from nbconvert.preprocessors import (
    ClearMetadataPreprocessor,
    ClearOutputPreprocessor,
    ExecutePreprocessor,
)

# import subprocess

# %%
nbdirs = [
    # os.path.join("docs/00tutorials"),
    os.path.join("docs/02examples"),
    # os.path.join("notebooks"),
]


# testdir = tempfile.mkdtemp()


def get_notebooks():
    nblist = []
    for nbdir in nbdirs:
        nblist += [
            os.path.join(nbdir, f) for f in os.listdir(nbdir) if f.endswith(".ipynb")
        ]
    return nblist


# def get_jupyter_kernel():
#     try:
#         jklcmd = ("jupyter", "kernelspec", "list")
#         b = subprocess.Popen(
#             jklcmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
#         ).communicate()[0]
#         if isinstance(b, bytes):
#             b = b.decode("utf-8")
#         print(b)
#         for line in b.splitlines():
#             if "python" in line:
#                 kernel = line.split()[0]
#     except:  # noqa
#         kernel = None

#     return kernel


# @pytest.mark.notebooks
# @pytest.mark.parametrize("pth", get_notebooks())
# def test_notebook(pth):
#     kernel = get_jupyter_kernel()
#     print("available jupyter kernel {}".format(kernel))

#     fn = os.path.basename(pth)

#     cmd = (
#         "jupyter "
#         + "nbconvert "
#         + "--ExecutePreprocessor.timeout=600 "
#         + "--to "
#         + "notebook "
#         + "--execute "
#         + "{} ".format(pth)
#         + "--output-dir "
#         + "{} ".format(testdir)
#         + "--output "
#         + "{}".format(fn)
#     )
#     ival = os.system(cmd)
#     assert ival == 0, "could not run {}".format(fn)


# @pytest.mark.notebooks
# @pytest.mark.parametrize("pth", get_notebooks())
# def test_notebook_cmd(pth):
#     cmd = (
#         "jupyter "
#         + "nbconvert "
#         + "--ExecutePreprocessor.timeout=600 "
#         + "--ExecutePreprocessor.kernel_name=python3 "
#         + "--to "
#         + "notebook "
#         + "--execute "
#         + "{} ".format(pth)
#     )
#     ival = os.system(cmd)
#     assert ival == 0, f"could not run {os.path.basename(pth)}"


@pytest.mark.notebooks
@pytest.mark.parametrize("pth", get_notebooks())
def test_notebook_py(pth):
    with open(pth, "r", encoding="utf-8") as f:
        nb = nbformat.read(f, as_version=4)
        ep = ExecutePreprocessor(timeout=600, kernel_name="python3")
        try:
            assert (
                ep.preprocess(nb, {"metadata": {"path": "docs/02examples"}}) is not None
            ), f"Got empty notebook for {os.path.basename(pth)}"
        except Exception as e:
            pytest.fail(reason=f"Failed executing {os.path.basename(pth)}: {e}")


# %% clear output and metadata of all notebooks
if __name__ == "__main__":
    clear_output = ClearOutputPreprocessor()
    clear_metadata = ClearMetadataPreprocessor()

    for notebook in get_notebooks():
        print("Clearing notebook:", notebook)
        with open(notebook, "r", encoding="utf-8") as f:
            nb = nbformat.read(f, as_version=4)

        # run nbconvert preprocessors to clear outputs and metadata
        clear_output.preprocess(nb, {})
        clear_metadata.preprocess(nb, {})

        with open(notebook, "w", encoding="utf-8") as f:
            nbformat.write(nb, f)



import numba
import numpy as np

# defenition of parameters

tiny = 1e-10
c = np.log(0.5) + 0.577215664901532860

fac = 1.0

nrange = np.arange(21, dtype=np.float64)

a = np.zeros(21, dtype=np.float64)
a[0] = 1.0
b = np.zeros(21, dtype=np.float64)

for n in range(1, 21):
    fac = n * fac
    a[n] = 1.0 / (4.0 ** nrange[n] * fac**2)
    b[n] = b[n - 1] + 1 / nrange[n]

b = (b - c) * a
a = -a / 2.0

gam = np.zeros((21, 21), dtype=np.float64)
for n in range(21):
    for m in range(n + 1):
        gam[n, m] = np.prod(nrange[m + 1 : n + 1]) / np.prod(nrange[1 : n - m + 1])

# gotta predefine these i.o. gam which is used in the old code
binom = np.zeros((21, 21), dtype=np.float64)
for n in range(21):
    for m in range(n + 1):
        binom[n, m] = np.prod(nrange[m + 1 : n + 1]) / np.prod(nrange[1 : n - m + 1])

# coefficients K1
fac = 1.0
bot = np.zeros(21, dtype=np.float64)
bot[0] = 4.0
for n in range(1, 21):
    fac = n * fac
    bot[n] = fac * (n + 1) * fac * 4.0 ** (n + 1)

psi = np.zeros(21, dtype=np.float64)
for n in range(2, 22):
    psi[n - 1] = psi[n - 2] + 1 / (n - 1)
psi = psi - 0.577215664901532860

a1 = np.empty(21, dtype=np.float64)
b1 = np.empty(21, dtype=np.float64)
twologhalf = 2 * np.log(0.5)
for n in range(21):
    a1[n] = 1 / bot[n]
    b1[n] = (twologhalf - (2.0 * psi[n] + 1 / (n + 1))) / bot[n]

# Laplace line doublet elements


@numba.njit(nogil=True, cache=True)
def lapld_int_ho(x, y, z1, z2, order):
    """lapld_int_ho.

    ! Near field only
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y
    complex(kind=8), intent(in) :: z1,z2
    complex(kind=8), dimension(0:order) :: omega, qm
    integer :: m, n
    real(kind=8) :: L
    complex(kind=8) :: z, zplus1, zmin1
    """
    omega = np.zeros(order + 1, dtype=np.complex128)
    qm = np.zeros(order + 1, dtype=np.complex128)

    # L = np.abs(z2 - z1)
    z = (2.0 * complex(x, y) - (z1 + z2)) / (z2 - z1)
    zplus1 = z + 1.0
    zmin1 = z - 1.0
    # Not sure if this gives correct answer at corner point (z also appears in qm);
    # should really be caught in code that calls this function
    if np.abs(zplus1) < tiny:
        zplus1 = tiny
    if np.abs(zmin1) < tiny:
        zmin1 = tiny

    omega[0] = np.log(zmin1 / zplus1)
    for n in range(1, order + 1):
        omega[n] = z * omega[n - 1]

    if order > 0:
        qm[1] = 2.0
    for m in range(3, order + 1, 2):
        qm[m] = qm[m - 2] * z * z + 2.0 / m

    for m in range(2, order + 1, 2):
        qm[m] = qm[m - 1] * z

    omega = 1.0 / (complex(0.0, 2.0) * np.pi) * (omega + qm)
    return omega


@numba.njit(nogil=True, cache=True)
def lapld_int_ho_d1d2(x, y, z1, z2, order, d1, d2):
    """lapld_int_ho_d1d2.

    Near field only
    Returns integral from d1 to d2 along real axis while strength is still
    Delta^order from -1 to +1
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,d1,d2
    complex(kind=8), intent(in) :: z1,z2
    complex(kind=8), dimension(0:order) :: omega, omegac
    integer :: n, m
    real(kind=8) :: xp, yp, dc, fac
    complex(kind=8) :: z1p,z2p,bigz1,bigz2
    """
    omega = np.zeros(order + 1, dtype=np.complex128)

    bigz1 = complex(d1, 0.0)
    bigz2 = complex(d2, 0.0)
    z1p = 0.5 * (z2 - z1) * bigz1 + 0.5 * (z1 + z2)
    z2p = 0.5 * (z2 - z1) * bigz2 + 0.5 * (z1 + z2)
    omegac = lapld_int_ho(x, y, z1p, z2p, order)
    dc = (d1 + d2) / (d2 - d1)
    for n in range(order + 1):
        for m in range(n + 1):
            omega[n] = omega[n] + gam[n, m] * dc ** (n - m) * omegac[m]
        omega[n] = (0.5 * (d2 - d1)) ** n * omega[n]

    return omega


@numba.njit(nogil=True, cache=True)
def lapld_int_ho_wdis(x, y, z1, z2, order):
    """lapld_int_ho_wdis.

    # Near field only
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y
    complex(kind=8), intent(in) :: z1,z2
    complex(kind=8), dimension(0:order) :: wdis
    complex(kind=8), dimension(0:10) :: qm  # Max order is 10
    integer :: m, n
    complex(kind=8) :: z, zplus1, zmin1, term1, term2, zterm
    """
    qm = np.zeros(11, dtype=np.complex128)
    wdis = np.zeros(order + 1, dtype=np.complex128)

    z = (2.0 * complex(x, y) - (z1 + z2)) / (z2 - z1)
    zplus1 = z + 1.0
    zmin1 = z - 1.0
    # Not sure if this gives correct answer at corner point (z also appears in qm);
    # should really be caught in code that calls this function
    if np.abs(zplus1) < tiny:
        zplus1 = tiny
    if np.abs(zmin1) < tiny:
        zmin1 = tiny

    qm[0:1] = 0.0
    for m in range(2, order + 1):
        qm[m] = 0.0
        for n in range(1, m // 2 + 1):
            qm[m] = qm[m] + (m - 2 * n + 1) * z ** (m - 2 * n) / (2 * n - 1)

    term1 = 1.0 / zmin1 - 1.0 / zplus1
    term2 = np.log(zmin1 / zplus1)
    wdis[0] = term1
    zterm = complex(1.0, 0.0)
    for m in range(1, order + 1):
        wdis[m] = m * zterm * term2 + z * zterm * term1 + 2.0 * qm[m]
        zterm = zterm * z

    wdis = -wdis / (np.pi * complex(0.0, 1.0) * (z2 - z1))
    return wdis


@numba.njit(nogil=True, cache=True)
def lapld_int_ho_wdis_d1d2(x, y, z1, z2, order, d1, d2):
    """lapld_int_ho_wdis_d1d2.

    # Near field only
    # Returns integral from d1 to d2 along real axis while strength is still
    # Delta^order from -1 to +1
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,d1,d2
    complex(kind=8), intent(in) :: z1,z2
    complex(kind=8), dimension(0:order) :: wdis, wdisc
    integer :: n, m
    real(kind=8) :: xp, yp, dc, fac
    complex(kind=8) :: z1p,z2p,bigz1,bigz2
    """
    wdis = np.zeros(order + 1, dtype=np.complex128)

    bigz1 = complex(d1, 0.0)
    bigz2 = complex(d2, 0.0)
    z1p = 0.5 * (z2 - z1) * bigz1 + 0.5 * (z1 + z2)
    z2p = 0.5 * (z2 - z1) * bigz2 + 0.5 * (z1 + z2)
    wdisc = lapld_int_ho_wdis(x, y, z1p, z2p, order)
    dc = (d1 + d2) / (d2 - d1)
    wdis[0 : order + 1] = 0.0
    for n in range(order + 1):
        for m in range(n + 1):
            wdis[n] = wdis[n] + gam[n, m] * dc ** (n - m) * wdisc[m]
        wdis[n] = (0.5 * (d2 - d1)) ** n * wdis[n]
    return wdis


# Fp function


@numba.njit(nogil=True, cache=True)
def Fp(x, y, z1, z2, biga, order, d1, d2, a, b, nt):
    tol = 1e-12
    zeta = (2 * complex(x, y) - (z1 + z2)) / (z2 - z1) / biga
    zetabar = np.conj(zeta)
    zminzbar = np.zeros(nt + 1, dtype=np.complex128)
    zminzbar[0] = 1
    for n in range(1, nt + 1):
        zminzbar[n] = zminzbar[n - 1] * (zeta - zetabar)

    eta = np.zeros((nt + 1, nt + 1), dtype=np.complex128)  # lower triangular
    etabar = np.zeros((nt + 1, nt + 1), dtype=np.complex128)
    for n in range(nt + 1):
        for m in range(0, n + 1):
            eta[n, m] = binom[n, m] * zminzbar[n - m]
            etabar[n, m] = np.conj(eta[n, m])

    atil = np.zeros(2 * nt + 1, dtype=np.complex128)
    btil = np.zeros(2 * nt + 1, dtype=np.complex128)
    ctil = np.zeros(2 * nt + 1, dtype=np.complex128)
    for n in range(2 * nt + 1):
        for m in range(max(0, n - nt), int(n / 2) + 1):
            atil[n] = atil[n] + a[n - m] * eta[n - m, m]
            btil[n] = btil[n] + b[n - m] * eta[n - m, m]
            ctil[n] = ctil[n] + a[n - m] * etabar[n - m, m]

    d1minzeta = d1 / biga - zeta
    d2minzeta = d2 / biga - zeta
    if np.abs(d1minzeta) < tol:
        d1minzeta = d1minzeta + complex(tol, 0)
    if np.abs(d2minzeta) < tol:
        d2minzeta = d2minzeta + complex(tol, 0)
    log1 = np.log(d1minzeta)
    log2 = np.log(d2minzeta)

    alpha = np.zeros(2 * nt + order + 1, dtype=np.complex128)
    beta = np.zeros(2 * nt + order + 1, dtype=np.complex128)
    gamma = np.zeros(2 * nt + order + 1, dtype=np.complex128)

    omega = np.zeros(order + 1, dtype=np.complex128)

    for p in range(order + 1):
        alpha[0 : 2 * nt + p + 1] = 0
        beta[0 : 2 * nt + p + 1] = 0
        gamma[0 : 2 * nt + p + 1] = 0

        d = np.zeros(p + 1, dtype=np.complex128)
        dbar = np.zeros(p + 1, dtype=np.complex128)
        for m in range(p + 1):
            d[m] = biga**p * binom[p, m] * zeta ** (p - m)
            dbar[m] = np.conj(d[m])
        for n in range(2 * nt + p + 1):
            for m in range(max(0, n - 2 * nt), min(p, n) + 1):
                alpha[n] = alpha[n] + d[m] * atil[n - m]
                beta[n] = beta[n] + d[m] * btil[n - m]
                gamma[n] = gamma[n] + dbar[m] * ctil[n - m]

        term1 = 1
        term2 = 1
        for n in range(2 * nt + p + 1):
            term1 = term1 * d1minzeta
            term2 = term2 * d2minzeta
            omega[p] = omega[p] + (
                alpha[n] * log2 - alpha[n] / (n + 1) + beta[n]
            ) * term2 / (n + 1)
            omega[p] = omega[p] - (
                alpha[n] * log1 - alpha[n] / (n + 1) + beta[n]
            ) * term1 / (n + 1)
            omega[p] = omega[p] + (
                gamma[n] * np.conj(log2) - gamma[n] / (n + 1)
            ) * np.conj(term2) / (n + 1)
            omega[p] = omega[p] - (
                gamma[n] * np.conj(log1) - gamma[n] / (n + 1)
            ) * np.conj(term1) / (n + 1)

    return biga * omega


# Bessel line elements


@numba.njit(nogil=True, cache=True)
def bessells_int_ho_new(x, y, z1, z2, lab, order, d1, d2, nt=20):
    """
    Docs.

    To come here
    """
    L = np.abs(z2 - z1)
    ang = np.arctan2(lab.imag, lab.real)
    biga = 2 * np.abs(lab) / L

    exprange = np.exp(-complex(0, 2) * ang * nrange)
    ahat = a * exprange
    bhat = (b - a * complex(0, 2) * ang) * exprange

    omega = Fp(x, y, z1, z2, biga, order, d1, d2, ahat, bhat, nt)
    return -L / (4 * np.pi) * omega


@numba.njit(nogil=True, cache=True)
def bessells_int_ho_qxqy_new(x, y, z1, z2, lab, order, d1, d2):
    """
    Docs.

    To come here
    """
    nt = 20  # number of terms in series is nt + 1
    bigz = (2 * complex(x, y) - (z1 + z2)) / (z2 - z1)
    bigx = bigz.real
    bigy = bigz.imag
    L = np.abs(z2 - z1)
    ang = np.arctan2(lab.imag, lab.real)
    angz = np.arctan2((z2 - z1).imag, (z2 - z1).real)
    biglab = 2 * lab / L
    biga = np.abs(biglab)

    exprange = np.exp(-complex(0, 2) * ang * nrange)
    ahat = a * exprange
    bhat = (b - a * complex(0, 2) * ang) * exprange

    atil = 2 * nrange[1:] * ahat[1:]
    btil = 2 * nrange[1:] * bhat[1:] + 2 * ahat[1:]

    omega = Fp(x, y, z1, z2, biga, order + 1, d1, d2, atil, btil, nt - 1)
    omegalap = lapld_int_ho_d1d2(x, y, z1, z2, order, d1, d2)
    term1 = 1 / (2 * np.pi * biga**2) * bigx * omega[:-1]
    term2 = -1 / (2 * np.pi * biga**2) * omega[1:]
    term3 = 2 * ahat[0] * omegalap.imag
    qx = term1 + term2 + term3
    term1 = 1 / (2 * np.pi * biga**2) * bigy * omega[:-1]
    term3 = 2 * ahat[0] * omegalap.real
    qy = term1 + term3

    qxqy = np.zeros(2 * order + 2, dtype=np.complex128)
    qxqy[: order + 1] = qx * np.cos(angz) - qy * np.sin(angz)
    qxqy[order + 1 :] = qx * np.sin(angz) + qy * np.cos(angz)
    return qxqy


@numba.njit(nogil=True, cache=True)
def besselld_int_ho_new(x, y, z1, z2, lab, order, d1, d2):
    """
    Docs.

    To come here
    """
    nt = 20  # number of terms in series is nt + 1
    bigz = (2 * complex(x, y) - (z1 + z2)) / (z2 - z1)
    bigy = bigz.imag
    L = np.abs(z2 - z1)
    ang = np.arctan2(lab.imag, lab.real)
    biglab = 2 * lab / L
    biga = np.abs(biglab)

    exprange = np.exp(-complex(0, 2) * ang * nrange)
    ahat = a1 * exprange
    bhat = (b1 - a1 * complex(0, 2) * ang) * exprange

    omega = Fp(x, y, z1, z2, biga, order, d1, d2, ahat, bhat, nt)

    rv = (
        bigy / (2.0 * np.pi * biglab**2) * omega
        + lapld_int_ho_d1d2(x, y, z1, z2, order, d1, d2).real
    )

    return rv


@numba.njit(nogil=True, cache=True)
def besselld_int_ho_qxqy_new(x, y, z1, z2, lab, order, d1, d2):
    """
    Docs.

    To come here
    """
    nt = 20  # number of terms in series is nt + 1
    bigz = (2 * complex(x, y) - (z1 + z2)) / (z2 - z1)
    bigx = bigz.real
    bigy = bigz.imag
    L = np.abs(z2 - z1)
    ang = np.arctan2(lab.imag, lab.real)
    angz = np.arctan2((z2 - z1).imag, (z2 - z1).real)
    biglab = 2 * lab / L
    biga = np.abs(biglab)

    exprange = np.exp(-complex(0, 2) * ang * nrange)
    ahat = a1 * exprange
    bhat = (b1 - a1 * complex(0, 2) * ang) * exprange

    atil = 2 * nrange[1:] * ahat[1:]
    btil = 2 * nrange[1:] * bhat[1:] + 2 * ahat[1:]

    omega_pot = Fp(x, y, z1, z2, biga, order, d1, d2, ahat, bhat, nt)
    omega = Fp(x, y, z1, z2, biga, order + 1, d1, d2, atil, btil, nt - 1)
    omegalap = lapld_int_ho_d1d2(x, y, z1, z2, order, d1, d2)
    wlap = lapld_int_ho_wdis_d1d2(x, y, z1, z2, order, d1, d2)

    term1 = bigx / (2 * np.pi * biga**2) * omega[:-1]
    term2 = -1 / (2 * np.pi * biga**2) * omega[1:]
    term3 = 2 * ahat[0] * omegalap.imag
    qx = -2 * bigy / (L * biglab**2) * (term1 + term2 + term3)  # + wlap.real

    term1 = 1 / (2.0 * np.pi * biglab**2) * 2 / L * omega_pot
    term2 = bigy / (2 * np.pi * biga**2) * omega[:-1]
    term3 = 2 * ahat[0] * omegalap.real
    qy = -term1 - 2 * bigy / (L * biglab**2) * (term2 + term3)  # - wlap.imag

    qxqy = np.zeros(2 * order + 2, dtype=np.complex128)
    qxqy[: order + 1] = qx * np.cos(angz) - qy * np.sin(angz) + wlap.real
    qxqy[order + 1 :] = qx * np.sin(angz) + qy * np.cos(angz) - wlap.imag
    return qxqy



import numpy as np


def timtrace(
    ml,
    xstart,
    ystart,
    zstart,
    tstartend,
    tstartoffset,
    tstep,
    nstepmax=100,
    hstepmax=10,
    silent=False,
    correctionstep=True,
):
    """Compute a pathline by numerical integration of the velocity vector.

    Pathline is broken up in sections for which starting times are provided.
    Pathline is computed from first starting time + offset until second
    starting time, then continued from second starting time + offset until
    third starting time, etc.

    Parameters
    ----------
    model : Model object
        model
    xstart : float
        x-coordinate of starting location of pathline
    ystart : float
        y-coordinate of starting location of pathline
    zstart : float
        z-coordinate of starting location of pathline
    tstartend : list
        list of starting times of pathline. last entry is the ending time.
    tstartoffset : float or list
        time after starting time when pathline is started. value or list. if
        this value is smaller than tmin it may cause problems after change in
        boundary conditions
    tstep : scalar or list
        maximum time step for each step along pathline. Either one value for
        all sections, or list with values for each section.
    nstepmax : integer
        maximum number of steps per section
    hstepmax : float or integer
        maximum length of horizontal step
    silent : boolean
        parameter to indicate if message should be printed to the screen for
        each section of the pathline
    correctionstep : boolean
        parameter to indicate if a correction step (Euler's method) should
        be taken. Taking a correction step is more accurate, especially for
        curved pathlines.

    Returns
    -------
    result : dictionary with three items
        * xyzt : 2D array with four columns: x, y, z, t along pathline
        * message : list with text messages of each section of the pathline
        * status : numerical indication of the result. Negative is likely
          undesirable.
            * -2 : reached maximum number of steps before reaching maximum time
            * -1 : starting z value not inside aquifer
            * +1 : reached maximum time
            * +2 : reached element
            * +3 : flows out of top of aquifer
    """
    xyzt = [np.array([[xstart, ystart, zstart, tstartend[0]]])]
    messages = []
    status = []
    if np.isscalar(tstep):
        tstep = len(tstartend) * [tstep]
    if np.isscalar(tstartoffset):
        tstartoffset = len(tstartend) * [tstartoffset]
    for itrace in range(len(tstartend) - 1):
        x0, y0, z0, t0 = xyzt[-1][-1]
        trace = timtraceline(
            ml,
            x0,
            y0,
            z0,
            t0 + tstartoffset[itrace],
            tstep[itrace],
            tstartend[itrace + 1],
            nstepmax=nstepmax,
            hstepmax=hstepmax,
            silent=silent,
            correctionstep=correctionstep,
        )
        xyzt.append(trace["xyzt"])
        messages.append(trace["message"])
        status.append(trace["status"])
        if trace["status"] != 1:
            break
    xyzt = np.vstack(xyzt)
    result = {"xyzt": np.array(xyzt), "message": messages, "status": status}
    return result


def timtraceline(
    ml,
    xstart,
    ystart,
    zstart,
    tstart,
    delt,
    tmax,
    nstepmax=100,
    hstepmax=10,
    correctionstep=True,
    silent=False,
):
    # treating aquifer layers and leaky layers the same way
    direction = 1  # forward
    terminate = False
    status = 0
    message = "no message"
    eps = 1e-6  # used to place point just above or below aquifer top or bottom
    aq = ml.aq.find_aquifer_data(xstart, ystart)
    if zstart > aq.z[0] or zstart < aq.z[-1]:
        terminate = True
        status = -1
        message = "starting z value not inside aquifer"
    # slightly alter starting location not to get stuck in surpring points
    # starting at time 0
    xyzt = [np.array([xstart * (1 + eps), ystart * (1 + eps), zstart, tstart])]
    layerlist = []  # to keep track of layers for plotting with colors
    for istep in range(nstepmax):
        if terminate:
            break
        do_correction = (
            correctionstep  # do correction step, unless do_correction changed to False
        )
        x0, y0, z0, t0 = xyzt[-1]
        # print(x0, y0, z0, t0)
        # aq = ml.aq.find_aquifer_data(x0, y0)  # find new aquifer
        layer, ltype, modellayer = aq.findlayer(z0)
        layerlist.append(modellayer)
        v0 = ml.velocomp(x0, y0, z0, t0, aq, [layer, ltype])
        vx, vy, vz = v0
        substep = 1  # take max 2 substeps

        for _ in range(2):
            if substep <= 2:
                # check if max time reached
                if t0 + delt > tmax:
                    delt0 = tmax - t0
                else:
                    delt0 = delt

                # check if horizontal step larger than hstepmax
                hstep = np.sqrt(vx**2 + vy**2) * delt0
                if hstep > hstepmax:
                    delt0 = hstepmax / hstep * delt0

                # check if going to different layer
                z1 = z0 + delt0 * vz
                layer1, ltype1, modellayer1 = aq.findlayer(z1)
                # print(steps, 'z1', z1, layer1, ltype1, modellayer1)
                if modellayer1 < modellayer:  # step up to next layer
                    delt0 = (aq.z[modellayer] - z0) / (z1 - z0) * delt0
                    z1 = aq.z[modellayer] + eps
                    # print('stepping up to next layer, z1= ', z1)
                    do_correction = False
                elif modellayer1 > modellayer:  # step down to next layer
                    delt0 = (z0 - aq.z[modellayer + 1]) / (z0 - z1) * delt0
                    z1 = aq.z[modellayer1] - eps
                    do_correction = False

                # potential new location
                x1 = x0 + delt0 * vx
                y1 = y0 + delt0 * vy
                t1 = t0 + delt0
                xyzt1 = np.array([x1, y1, z1, t1])

                # check elements if point needs to be changed
                for e in ml.elementlist:
                    changed, terminate, xyztnew, message = e.changetrace(
                        xyzt[-1],
                        xyzt1,
                        aq,
                        layer,
                        ltype,
                        modellayer,
                        direction,
                        hstepmax,
                    )
                    if changed or terminate:
                        x1, y1, z1, t1 = xyztnew
                        do_correction = False
                        status = 2
                        message = message
                        break

                if t1 >= tmax:
                    terminate = True
                    status = 1
                    message = "reached maximum time tmax"
                if istep == nstepmax - 1:
                    terminate = True
                    status = -2
                    message = "reached maximum number of steps"
                if z1 > aq.z[0]:
                    terminate = True
                    message = "flows out of top of aquifer"
                    status = 3
                if terminate:
                    xyzt.append(np.array([x1, y1, z1, t1]))
                    break

                if substep == 1 and do_correction:  # do correction step
                    vnew = ml.velocomp(x1, y1, z1, t1, aq, [layer, ltype])
                    v1 = 0.5 * (v0 + vnew)
                    vx, vy, vz = v1
                    substep = 2
                else:
                    xyzt.append(np.array([x1, y1, z1, t1]))
                    break

    if not silent:
        print(message)
    result = {"xyzt": np.array(xyzt), "message": message, "status": status}
    return result


# test with tmult. didn't improve much
def timtraceline2(
    ml,
    xstart,
    ystart,
    zstart,
    tstart,
    tmax,
    delt,
    deltmin,
    deltmax,
    tmult=1.1,
    nstepmax=100,
    hstepmax=10,
    silent=False,
    correct=False,
):
    # treating aquifer layers and leaky layers the same way
    direction = 1  # forward
    terminate = False
    message = "no message"
    eps = 1e-10  # used to place point just above or below aquifer top or bottom
    aq = ml.aq.find_aquifer_data(xstart, ystart)
    if zstart > aq.z[0] or zstart < aq.z[-1]:
        terminate = True
        message = "starting z value not inside aquifer"
    # slightly alter starting location not to get stuck in surpring points
    # starting at time 0
    xyzt = [np.array([xstart * (1 + eps), ystart * (1 + eps), zstart, tstart])]
    layerlist = []  # to keep track of layers for plotting with colors
    speednew = 0.0
    for istep in range(nstepmax):
        if terminate:
            break
        do_correction = correct  # do correction, unless do_correction changed to False
        x0, y0, z0, t0 = xyzt[-1]
        # aq = ml.aq.find_aquifer_data(x0, y0)  # find new aquifer
        layer, ltype, modellayer = aq.findlayer(z0)
        layerlist.append(modellayer)
        speedold = speednew
        v0 = ml.velocomp(x0, y0, z0, t0, aq, [layer, ltype])
        speednew = np.sqrt(np.sum(v0**2))
        if speednew < speedold:
            delt *= 1.2
            if delt > deltmax:
                delt = deltmax
        else:
            delt /= 1.2
            if delt < deltmin:
                delt = deltmin
        print("delt:", delt)
        vx, vy, vz = v0
        substep = 1  # take max 2 substeps

        for _ in range(2):
            if substep <= 2:
                # check if max time reached
                if t0 + delt > tmax:
                    delt0 = tmax - t0
                else:
                    delt0 = delt

                # check if horizontal step larger than hstepmax
                hstep = np.sqrt(vx**2 + vy**2) * delt0
                if hstep > hstepmax:
                    delt0 = hstepmax / hstep * delt0

                # check if going to different layer
                z1 = z0 + delt0 * vz
                layer1, ltype1, modellayer1 = aq.findlayer(z1)
                if modellayer1 < modellayer:  # step up to next layer
                    delt0 = (aq.z[modellayer] - z0) / (z1 - z0) * delt0
                    z1 = aq.z[modellayer] + eps
                    do_correction = False
                elif modellayer1 > modellayer:  # step down to next layer
                    delt0 = (z0 - aq.z[modellayer + 1]) / (z0 - z1) * delt0
                    z1 = aq.z[modellayer1] - eps
                    do_correction = False

                # potential new location
                x1 = x0 + delt0 * vx
                y1 = y0 + delt0 * vy
                t1 = t0 + delt0
                xyzt1 = np.array([x1, y1, z1, t1])

                # check elements if point needs to be changed
                for e in ml.elementlist:
                    changed, terminate, xyztnew, changemessage = e.changetrace(
                        xyzt[-1],
                        xyzt1,
                        aq,
                        layer,
                        ltype,
                        modellayer,
                        direction,
                        hstepmax,
                    )
                    if changed or terminate:
                        x1, y1, z1, t1 = xyztnew
                        do_correction = False
                        message = changemessage
                        break

                if t1 >= tmax:
                    terminate = True
                    message = "reached maximum time tmax"
                if istep == nstepmax - 1:
                    terminate = True
                    message = "reached maximum number of steps"
                if terminate:
                    xyzt.append(np.array([x1, y1, z1, t1]))
                    break

                if substep == 1 and do_correction:  # do correction step
                    vnew = ml.velocomp(x1, y1, z1, t1, aq, [layer, ltype])
                    v1 = 0.5 * (v0 + vnew)
                    vx, vy, vz = v1
                    substep = 2
                else:
                    xyzt.append(np.array([x1, y1, z1, t1]))
                    break

    if not silent:
        print(message)
    result = {"xyzt": np.array(xyzt), "message": message, "complete": terminate}
    return result



import re
import warnings
from typing import Iterable

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.optimize import least_squares, leastsq


class Calibrate:
    def __init__(self, model):
        """Initialize Calibration class.

        Parameters
        ----------
        model : ttim.Model
            model to calibrate
        """
        self.model = model
        self.parameters = pd.DataFrame(
            columns=[
                "layers",
                "optimal",
                "std",
                "perc_std",
                "pmin",
                "pmax",
                "initial",
                "inhoms",
                "parray",
            ]
        )
        self.seriesdict = {}
        self.seriesinwelldict = {}

    def set_parameter(
        self, name=None, layers=None, initial=0, pmin=-np.inf, pmax=np.inf, inhoms=None
    ):
        """Set parameter to be optimized.

        Parameters
        ----------
        name : str
            name can be 'kaq', 'Saq', 'c', 'Sll' or 'kzoverkh'.
        layers : int or list of ints
            layer number(s) for which the parameter is set. If an integer is passed,
            parameter is associated with a single layer. If a list of layers is passed,
            layers must be consecutive and parameter is set for each layer from
            min(layers) up to and including max(layers).
        initial : float, optional
            initial value for the parameter (the default is 0)
        pmin : float, optional
            lower bound for parameter value (the default is -np.inf)
        pmax : float, optional
            upper bound for paramater value (the default is np.inf)
        inhoms : str, list
            string with name of inhomogeneity, list with string names of inhomogeneities
            or list of inhomogeneities
            inhomogeneity(ies) for which the parameter is set. If a string is passed,
            parameter is associated with a single inhomogeneity. If a list of strings
            of inhomogeneities (or list of inhomogeneities) is passed, parameter is set
            for each inhomogeneity in the list. This allows linking of parameters across
            inhomogeneities.
        """
        assert isinstance(name, str), "Error: name must be string"

        if isinstance(layers, Iterable):
            from_lay = min(layers)
            to_lay = max(layers)
            if (np.diff(layers) > 1).any():
                warnings.warn(
                    "Non-consecutive layers are not supported. "
                    f"Setting parameter '{name}' for layers {from_lay} - {to_lay}.",
                    stacklevel=1,
                )
        elif isinstance(layers, int):
            from_lay = layers
            to_lay = layers
        else:
            warnings.warn(
                "Setting layers in the parameter name is deprecated. "
                f"Set the layers= keyword argument for parameter '{name}' to silence "
                "this warning. The parameter name can still include layer info, but "
                "this will be ignored in a future version of TTim.",
                DeprecationWarning,
                stacklevel=2,
            )
            # find numbers in name str for support layer ranges
            layers_from_name = re.findall(r"\d+", name)
            if len(layers_from_name) == 0:
                raise ValueError(
                    "No layer information found in parameter name. "
                    "Please specify layers explicitly."
                )
            elif len(layers_from_name) == 1:
                from_lay = int(layers_from_name[0])
                to_lay = from_lay
            elif len(layers_from_name) == 2:
                from_lay, to_lay = map(int, layers_from_name)

        # get aquifer information and create list if necessary
        if inhoms is None:
            aq = [self.model.aq]
        elif isinstance(inhoms, tuple):
            aq = list(inhoms)
        elif not isinstance(inhoms, list):
            aq = [inhoms]
        else:
            aq = inhoms

        # convert aquifer names to aquifer objects
        for i, iaq in enumerate(aq):
            if isinstance(iaq, str):
                aq[i] = self.model.aq.inhomdict[iaq]

        plist = []
        for iaq in aq:
            if name[:3] == "kaq":
                p = iaq.kaq[from_lay : to_lay + 1]
            elif name[:3] == "Saq":
                p = iaq.Saq[from_lay : to_lay + 1]
            elif name[0] == "c":
                p = iaq.c[from_lay : to_lay + 1]
            elif name[:3] == "Sll":
                p = iaq.Sll[from_lay : to_lay + 1]
            elif name[0:8] == "kzoverkh":
                p = iaq.kzoverkh[from_lay : to_lay + 1]
            else:
                raise ValueError(
                    f"Parameter name '{name}' not recognized. "
                    "Supported parameters are 'kaq', 'Saq', 'c', 'Sll' or 'kzoverkh'."
                )
            plist.append(p[:])

        if p is None:  # no parameter set
            print("parameter name not recognized or no parameter ref supplied")
            return

        if inhoms is None:
            pname = f"{name}_{from_lay}_{to_lay}"
        else:
            pname = f"{name}_{from_lay}_{to_lay}_{'_'.join([iaq.name for iaq in aq])}"
        self.parameters.loc[pname] = {
            "layers": layers,
            "optimal": float(initial),
            "std": None,
            "perc_std": None,
            "pmin": float(pmin),
            "pmax": float(pmax),
            "initial": float(initial),
            "inhoms": aq if inhoms is not None else None,
            "parray": plist,
        }

    def set_parameter_by_reference(
        self, name=None, parameter=None, initial=0, pmin=-np.inf, pmax=np.inf
    ):
        """Set parameter to be optimized.

        Parameters
        ----------
        name : str
            parameter name
        parameter : np.array
            array reference containing the parameter to be optimized. must be
            specified as reference, i.e. w.rc[0:]
        initial : float, optional
            initial value for the parameter (the default is 0)
        pmin : float, optional
            lower bound for parameter value (the default is -np.inf)
        pmax : float, optional
            upper bound for paramater value (the default is np.inf)
        """
        assert isinstance(name, str), "Error: name must be string"
        if parameter is not None:
            assert isinstance(parameter, np.ndarray), (
                "Error: parameter needs to be numpy array"
            )
            p = parameter
        self.parameters.loc[name] = {
            "optimal": float(initial),
            "std": None,
            "perc_std": None,
            "pmin": float(pmin),
            "pmax": float(pmax),
            "initial": float(initial),
            "parray": [p[:]],
        }

    def series(self, name, x, y, layer, t, h, weights=None):
        """Method to add observations to Calibration object.

        Parameters
        ----------
        name : str
            name of series
        x : float
            x-coordinate
        y : float
            y-coordinate
        layer : int
            layer number, 0-indexed
        t : np.array
            array containing timestamps of timeseries
        h : np.array
            array containing timeseries values, i.e. head observations
        """
        s = Series(x, y, layer, t, h, weights=weights)
        self.seriesdict[name] = s

    def seriesinwell(self, name, element, t, h):
        """Method to add observations to Calibration object.

        Parameters
        ----------
        name : str
            name of series
        element: element object with headinside function
        t : np.array
            array containing timestamps of timeseries
        h : np.array
            array containing timeseries values, i.e. head observations
        """
        e = SeriesInWell(element, t, h)
        self.seriesinwelldict[name] = e

    def residuals(self, p, printdot=False, weighted=True, layers=None, series=None):
        """Method to calculate residuals given certain parameters.

        Parameters
        ----------
        p : np.array
            array containing parameter values
        printdot : bool, optional
            print dot for each function call

        Returns
        -------
        np.array
            array containing all residuals
        """
        if printdot:
            print(".", end="")
        # set the values of the variables

        if printdot == 7:
            print(p)

        if layers is None:
            layers = range(self.model.aq.naq)

        for i, k in enumerate(self.parameters.index):
            parraylist = self.parameters.loc[k, "parray"]
            for parray in parraylist:
                # [:] needed to do set value in array
                parray[:] = p[i]
        self.model.solve(silent=True)

        rv = np.empty(0)
        cal_series = self.seriesdict.keys() if series is None else series
        for key in cal_series:
            s = self.seriesdict[key]
            if s.layer not in layers:
                continue
            h = self.model.head(s.x, s.y, s.t, layers=s.layer)
            w = s.weights if ((s.weights is not None) and weighted) else np.ones_like(h)
            rv = np.append(rv, (s.h - h) * w)
        for key in self.seriesinwelldict:
            s = self.seriesinwelldict[key]
            h = s.element.headinside(s.t)[0]
            rv = np.append(rv, s.h - h)
        return rv

    def residuals_lmfit(self, lmfitparams, printdot=False):
        vals = lmfitparams.valuesdict()
        p = np.array([vals[k] for k in self.parameters.index])
        return self.residuals(p, printdot)

    def fit_least_squares(self, report=True, diff_step=1e-4, xtol=1e-8, method="lm"):
        self.fitresult = least_squares(
            self.residuals,
            self.parameters.initial.values,
            args=(True,),
            bounds=(self.parameters.pmin.values, self.parameters.pmax.values),
            method=method,
            diff_step=diff_step,
            xtol=xtol,
            x_scale="jac",
        )
        print("", flush=True)
        # Call residuals to specify optimal values for model
        res = self.residuals(self.fitresult.x)
        for ipar in self.parameters.index:
            self.parameters.loc[ipar, "optimal"] = self.parameters.loc[ipar, "parray"][0]
        nparam = len(self.fitresult.x)
        H = self.fitresult.jac.T @ self.fitresult.jac
        sigsq = np.var(res, ddof=nparam)
        self.covmat = np.linalg.inv(H) * sigsq
        self.sig = np.sqrt(np.diag(self.covmat))
        D = np.diag(1 / self.sig)
        self.cormat = D @ self.covmat @ D
        self.parameters["std"] = self.sig
        self.parameters["perc_std"] = self.sig / self.parameters["optimal"] * 100
        if report:
            print(self.parameters)
            print(self.sig)
            print(self.covmat)
            print(self.cormat)

    def fit_lmfit(self, report=False, printdot=True, **kwargs):
        import lmfit

        self.lmfitparams = lmfit.Parameters()
        for name in self.parameters.index:
            p = self.parameters.loc[name]
            self.lmfitparams.add(name, value=p["initial"], min=p["pmin"], max=p["pmax"])
        fit_kws = {"epsfcn": 1e-4}  # this is essential to specify step for the Jacobian
        self.fitresult = lmfit.minimize(
            self.residuals_lmfit,
            self.lmfitparams,
            method="leastsq",
            kws={"printdot": printdot},
            **fit_kws,
            **kwargs,
        )
        print("", flush=True)
        print(self.fitresult.message)
        if self.fitresult.success:
            for name in self.parameters.index:
                self.parameters.loc[name, "optimal"] = self.fitresult.params.valuesdict()[
                    name
                ]
            if hasattr(self.fitresult, "covar"):
                self.parameters["std"] = np.sqrt(np.diag(self.fitresult.covar))
                self.parameters["perc_std"] = (
                    100 * self.parameters["std"] / np.abs(self.parameters["optimal"])
                )
            else:
                self.parameters["std"] = np.nan
                self.parameters["perc_std"] = np.nan
        if report:
            print(lmfit.fit_report(self.fitresult))

    def residuals_leastsq(self, logparams, printdot=False):
        params = 10**logparams
        print("params ", params)
        return self.residuals(params, printdot)

    def fit_leastsq(self, report=True, diff_step=1e-4, xtol=1e-8):
        params_initial = np.log10(self.parameters.initial.values)
        print("params_initial ", params_initial)
        plog, mes = leastsq(self.residuals_leastsq, params_initial, epsfcn=1e-3)
        print("", flush=True)
        params = 10**plog
        # Call residuals to specify optimal values for model
        self.residuals(params)
        for ipar in self.parameters.index:
            self.parameters.loc[ipar, "optimal"] = self.parameters.loc[ipar, "parray"][0]
        # nparam = len(self.fitresult.x)
        # H = self.fitresult.jac.T @ self.fitresult.jac
        # sigsq = np.var(res, ddof=nparam)
        # self.covmat = np.linalg.inv(H) * sigsq
        # self.sig = np.sqrt(np.diag(self.covmat))
        # D = np.diag(1 / self.sig)
        # self.cormat = D @ self.covmat @ D
        # self.parameters["std"] = self.sig
        # self.parameters["perc_std"] = self.sig / self.parameters["optimal"] * 100
        # if report:
        #     print(self.parameters)
        #     print(self.sig)
        #     print(self.covmat)
        #     print(self.cormat)

    def fit(self, report=False, printdot=True, **kwargs):
        # current default fitting routine is lmfit
        # return self.fit_least_squares(report) # does not support bounds by default
        return self.fit_lmfit(report, printdot, **kwargs)

    def rmse(self, weighted=True, layers=None):
        """Calculate root-mean-squared-error.

        Returns
        -------
        float
            return rmse value
        """
        r = self.residuals(
            self.parameters["optimal"].values, weighted=weighted, layers=layers
        )
        return np.sqrt(np.mean(r**2))

    def topview(self, ax=None, layers=None, labels=True):
        """Plot topview of model with calibration points.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            axes to plot on (the default is None, which creates a new figure)
        """
        if ax is None:
            _, ax = plt.subplots()
            # self.model.plots.topview(ax=ax)
        for key, s in self.seriesdict.items():
            if layers is None or s.layer in layers:
                ax.plot(s.x, s.y, "ko")
                if labels:
                    ax.text(s.x, s.y, key, ha="left", va="bottom")
        return ax

    def xsection(self, ax=None, labels=True):
        """Plot cross-section of model with calibration points.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            axes to plot on (the default is None, which creates a new figure)
        """
        if ax is None:
            _, ax = plt.subplots()
        for key, s in self.seriesdict.items():
            aq = self.model.aq.find_aquifer_data(s.x, s.y)
            ztop = aq.z[0]
            zpb_top = aq.zaqtop[s.layer]
            zpb_bot = aq.zaqbot[s.layer]
            ax.plot([s.x, s.x], [zpb_top, zpb_bot], c="k", ls="dotted")
            ax.plot([s.x, s.x], [ztop + 1, zpb_top], c="k", ls="solid", lw=1.0)
            if labels:
                ax.text(s.x, ztop + 1, key, ha="left", va="bottom", rotation=45)
        return ax


class Series:
    def __init__(self, x, y, layer, t, h, weights=None):
        self.x = x
        self.y = y
        self.layer = layer
        self.t = t
        self.h = h
        self.weights = weights


class SeriesInWell:
    def __init__(self, element, t, h):
        self.element = element
        self.t = t
        self.h = h



import inspect  # Used for storing the input

import matplotlib.pyplot as plt
import numpy as np
from scipy.special import iv, kv

from .element import Element


class CircAreaSink(Element):
    """Create a circular area-sink with uniform infiltration rate in aquifer layer 0.

    Infiltration rate in length / time, positive for water entering the aquifer.

    Parameters
    ----------
    model : Model object
        model to which the element is added
    xc : float
        x-coordinate of center of area-sink
    yc : float
        y-coordinate of center of area-sink
    R : radius of area-sink
    tsandN : list of tuples
        tuples of starting time and infiltration rate after starting time
    label : string or None (default: None)
        label of the area-sink
    """

    def __init__(
        self, model, xc=0, yc=0, R=0.1, tsandN=[(0, 1)], name="CircAreaSink", label=None
    ):
        self.storeinput(inspect.currentframe())
        Element.__init__(
            self,
            model,
            nparam=1,
            nunknowns=0,
            layers=0,
            tsandbc=tsandN,
            type="g",
            name=name,
            label=label,
        )
        self.xc = float(xc)
        self.yc = float(yc)
        self.R = float(R)
        self.model.addelement(self)

    def __repr__(self):
        return self.name + " at " + str((self.xc, self.yc))

    def initialize(self):
        self.aq = self.model.aq.find_aquifer_data(self.xc, self.yc)
        self.setbc()
        self.setflowcoef()
        # Since recharge is in layer 0, and RHS is -N
        self.an = self.aq.coef[0, :] * self.flowcoef
        self.an.shape = (self.aq.naq, self.model.nint, self.model.npint)
        self.termin = self.aq.lab2 * self.R * self.an
        self.termin2 = self.aq.lab2**2 * self.an
        self.terminq = self.R * self.an
        self.termout = self.aq.lab2 * self.R * self.an
        self.i1R = iv(1, self.R / self.aq.lab2)
        self.k1R = kv(1, self.R / self.aq.lab2)
        self.termoutq = self.R * self.an
        self.dischargeinf = self.aq.coef[0, :] * self.flowcoef
        self.dischargeinflayers = np.sum(
            self.dischargeinf * self.aq.eigvec[self.layers, :, :], 1
        )

    def setflowcoef(self):
        """Separate function so that this can be overloaded for other types."""
        self.flowcoef = 1.0 / self.model.p  # Step function

    def potinf(self, x, y, aq=None):
        """Can be called with only one x,y value."""
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros(
            (self.nparam, aq.naq, self.model.nint, self.model.npint), dtype=complex
        )
        if aq == self.aq:
            r = np.sqrt((x - self.xc) ** 2 + (y - self.yc) ** 2)
            # pot = np.zeros(self.model.npint, dtype=complex)
            if r < self.R:
                for i in range(self.aq.naq):
                    for j in range(self.model.nint):
                        # if r / abs(self.aq.lab2[i,j,0]) < self.rzero:
                        rv[0, i, j] = (
                            -self.termin[i, j] * self.K1RI0r(r, i, j) + self.termin2[i, j]
                        )
            else:
                for i in range(self.aq.naq):
                    for j in range(self.model.nint):
                        if (r - self.R) / abs(self.aq.lab2[i, j, 0]) < self.rzero:
                            rv[0, i, j, :] = self.termout[i, j, :] * self.I1RK0r(r, i, j)
        rv.shape = (self.nparam, aq.naq, self.model.npval)
        return rv

    def disvecinf(self, x, y, aq=None):
        """Can be called with only one x,y value."""
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        qx = np.zeros((self.nparam, aq.naq, self.model.npval), dtype=complex)
        qy = np.zeros((self.nparam, aq.naq, self.model.npval), dtype=complex)
        if aq == self.aq:
            qr = np.zeros(
                (self.nparam, aq.naq, self.model.nint, self.model.npint), dtype=complex
            )
            r = np.sqrt((x - self.xc) ** 2 + (y - self.yc) ** 2)
            if r < self.R:
                for i in range(self.aq.naq):
                    for j in range(self.model.nint):
                        # if r / abs(self.aq.lab2[i,j,0]) < self.rzero:
                        qr[0, i, j] = self.terminq[i, j] * self.K1RI1r(r, i, j)
            else:
                for i in range(self.aq.naq):
                    for j in range(self.model.nint):
                        if (r - self.R) / abs(self.aq.lab2[i, j, 0]) < self.rzero:
                            qr[0, i, j] = self.termoutq[i, j, :] * self.I1RK1r(r, i, j)
            qr.shape = (self.nparam, aq.naq, self.model.npval)
            qx[:] = qr * (x - self.xc) / r
            qy[:] = qr * (y - self.yc) / r
        return qx, qy

    def plot(self, ax=None):
        if ax is None:
            _, ax = plt.subplots()
        ax.plot(
            self.xc + self.R * np.cos(np.linspace(0, 2 * np.pi, 100)),
            self.yc + self.R * np.sin(np.linspace(0, 2 * np.pi, 100)),
            "k",
        )

    def K1RI0r(self, rin, iaq, ipint):
        r = rin / self.aq.lab2[iaq, ipint]
        R = self.R / self.aq.lab2[iaq, ipint]
        if np.isinf(self.i1R[iaq, ipint]).any():
            rv = (
                np.sqrt(1 / (4 * r * R))
                * np.exp(r - R)
                * (1 + 3 / (8 * R) - 15 / (128 * R**2) + 315 / (3072 * R**3))
                * (1 + 1 / (8 * r) + 9 / (128 * r**2) + 225 / (3072 * r**3))
            )
        else:
            rv = self.k1R[iaq, ipint] * iv(0, r)
        return rv

    def I1RK0r(self, rin, iaq, ipint):
        r = rin / self.aq.lab2[iaq, ipint]
        R = self.R / self.aq.lab2[iaq, ipint]
        if np.isinf(self.i1R[iaq, ipint]).any():
            rv = (
                np.sqrt(1 / (4 * r * R))
                * np.exp(R - r)
                * (1 - 3 / (8 * R) - 15 / (128 * R**2) - 315 / (3072 * R**3))
                * (1 - 1 / (8 * r) + 9 / (128 * r**2) - 225 / (3072 * r**3))
            )
        else:
            rv = self.i1R[iaq, ipint] * kv(0, r)
        return rv

    def K1RI1r(self, rin, iaq, ipint):
        r = rin / self.aq.lab2[iaq, ipint]
        R = self.R / self.aq.lab2[iaq, ipint]
        if np.isinf(self.i1R[iaq, ipint]).any():
            rv = (
                np.sqrt(1 / (4 * r * R))
                * np.exp(r - R)
                * (1 + 3 / (8 * R) - 15 / (128 * R**2) + 315 / (3072 * R**3))
                * (1 - 3 / (8 * r) - 15 / (128 * r**2) - 315 / (3072 * r**3))
            )
        else:
            rv = self.k1R[iaq, ipint] * iv(1, r)
        return rv

    def I1RK1r(self, rin, iaq, ipint):
        r = rin / self.aq.lab2[iaq, ipint]
        R = self.R / self.aq.lab2[iaq, ipint]
        if np.isinf(self.i1R[iaq, ipint]).any():
            rv = (
                np.sqrt(1 / (4 * r * R))
                * np.exp(R - r)
                * (1 - 3 / (8 * R) - 15 / (128 * R**2) - 315 / (3072 * R**3))
                * (1 + 3 / (8 * r) - 15 / (128 * r**2) + 315 / (3072 * r**3))
            )
        else:
            rv = self.i1R[iaq, ipint] * kv(1, r)
        return rv



import inspect  # Used for storing the input
from abc import ABC, abstractmethod

import numpy as np

from .invlapnumba import invlapcomp


class Element(ABC):
    def __init__(
        self,
        model,
        nparam=1,
        nunknowns=0,
        layers=0,
        tsandbc=[(0, 0)],
        type="z",
        name="",
        label=None,
        inhomelement=False,
    ):
        """Element base class.

        Types of elements:
        * 'g': strength is given through time
        * 'v': boundary condition is variable through time
        * 'z': boundary condition is zero through time

        Definition of nlayers, Ncp, Npar, nunknowns:
        * nlayers: Number of layers that the element is screened in,
          as set in Element
        * Ncp: Number of control points along the element
        * nparam: Number of parameters, commonly nlayers * Ncp
        * nunknowns: Number of unknown parameters, commonly zero or Npar
        """
        self.model = model
        self.aq = None  # Set in the initialization function
        self.nparam = nparam  # Number of parameters
        self.nunknowns = nunknowns
        self.layers = np.atleast_1d(layers)
        self.nlayers = len(self.layers)
        self.inhomelement = inhomelement  # tag inhomelements
        #
        tsandbc = np.atleast_2d(tsandbc).astype(float)
        tsandbc_error = (
            "tsandQ or tsandh need to be 2D lists"
            + " or arrays, like [(0, 1), (2, 5), (8, 0)] "
        )
        assert tsandbc.shape[1] == 2, tsandbc_error
        self.tstart, self.bcin = tsandbc[:, 0] - self.model.tstart, tsandbc[:, 1]
        if self.tstart[0] > 0:
            self.tstart = np.hstack((np.zeros(1), self.tstart))
            self.bcin = np.hstack((np.zeros(1), self.bcin))
        #
        # 'z' boundary condition thru time or 'v' boundary condition thru time
        self.type = type
        self.name = name
        self.label = label
        if self.label is not None:
            assert self.label not in self.model.elementdict.keys(), (
                "TTim error: label " + self.label + " already exists"
            )
        self.rzero = 30

    def setbc(self):
        if len(self.tstart) > 1:
            self.bc = np.zeros_like(self.bcin)
            self.bc[0] = self.bcin[0]
            self.bc[1:] = self.bcin[1:] - self.bcin[:-1]
        else:
            self.bc = self.bcin.copy()
        self.ntstart = len(self.tstart)

    @abstractmethod
    def initialize(self):
        """Initialize the element.

        Initialization of terms that cannot be initialized before other elements or
        the aquifer is defined.

        As we don't want to require a certain order of entering elements, these terms
        are initialized when Model.solve is called The initialization class needs to be
        overloaded by all derived classes
        """

    @abstractmethod
    def potinf(self, x, y, aq=None):
        """Returns complex array of size (nparam, naq, npval)."""

    def potential(self, x, y, aq=None):
        """Returns complex array of size (ngvbc, naq, npval)."""
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        return np.sum(
            self.parameters[:, :, np.newaxis, :] * self.potinf(x, y, aq), axis=1
        )

    def unitpotential(self, x, y, aq=None):
        """Returns complex array of size (naq, npval).

        Can be more efficient for given elements.
        """
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        return np.sum(self.potinf(x, y, aq), 0)

    def unitpotentialone(self, x, y, jtime, aq=None):
        """Returns complex array of size (naq, npval).

        Can be more efficient for given elements.
        """
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        return np.sum(self.potinfone(x, y, jtime, aq), 0)

    @abstractmethod
    def disvecinf(self, x, y, aq=None):
        """Returns 2 complex arrays of size (nparam, naq, npval)."""

    def disvec(self, x, y, aq=None):
        """Returns 2 complex arrays of size (ngvbc, naq, npval)."""
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        qx, qy = self.disvecinf(x, y, aq)
        return np.sum(self.parameters[:, :, np.newaxis, :] * qx, 1), np.sum(
            self.parameters[:, :, np.newaxis, :] * qy, 1
        )

    def unitdisvec(self, x, y, aq=None):
        """Returns 2 complex arrays of size (naq, npval).

        Can be more efficient for given elements.
        """
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        qx, qy = self.disvecinf(x, y, aq)
        return np.sum(qx, 0), np.sum(qy, 0)

    # Functions used to build equations
    def potinflayers(self, x, y, layers=0, aq=None):
        """Layers can be scalar, list, or array.

        returns array of size (len(layers),nparam,npval) only used in building equations
        """
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        pot = self.potinf(x, y, aq)
        rv = np.sum(pot[:, np.newaxis, :, :] * aq.eigvec, 2)
        # first axis needs to be the number of layers
        rv = rv.swapaxes(0, 1)
        return rv[layers, :]

    def potentiallayers(self, x, y, layers=0, aq=None):
        """Returns complex array of size (ngvbc, len(layers),npval).

        Only used in building equations.
        """
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        pot = self.potential(x, y, aq)
        phi = np.sum(pot[:, np.newaxis, :, :] * aq.eigvec, 2)
        return phi[:, layers, :]

    def unitpotentiallayers(self, x, y, layers=0, aq=None):
        """Returns complex array of size (len(layers), npval).

        Only used in building equations.
        """
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        pot = self.unitpotential(x, y, aq)
        phi = np.sum(pot[np.newaxis, :, :] * aq.eigvec, 1)
        return phi[layers, :]

    def disvecinflayers(self, x, y, layers=0, aq=None):
        """Layers can be scalar, list, or array.

        returns 2 arrays of size (len(layers),nparam,npval) only used in building
        equations
        """
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        qx, qy = self.disvecinf(x, y, aq)
        rvx = np.sum(qx[:, np.newaxis, :, :] * aq.eigvec, 2)
        rvy = np.sum(qy[:, np.newaxis, :, :] * aq.eigvec, 2)
        # first axis needs to be the number of layers
        rvx = rvx.swapaxes(0, 1)
        rvy = rvy.swapaxes(0, 1)
        return rvx[layers, :], rvy[layers, :]

    def disveclayers(self, x, y, layers=0, aq=None):
        """Returns 2 complex array of size (ngvbc, len(layers), npval).

        Only used in building equations.
        """
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        qx, qy = self.disvec(x, y, aq)
        rvx = np.sum(qx[:, np.newaxis, :, :] * aq.eigvec, 2)
        rvy = np.sum(qy[:, np.newaxis, :, :] * aq.eigvec, 2)
        return rvx[:, layers, :], rvy[:, layers, :]

    def unitdisveclayers(self, x, y, layers=0, aq=None):
        """Returns complex array of size (len(layers), npval).

        Only used in building equations.
        """
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        qx, qy = self.unitdisvec(x, y, aq)
        rvx = np.sum(qx[np.newaxis, :, :] * aq.eigvec, 1)
        rvy = np.sum(qy[np.newaxis, :, :] * aq.eigvec, 1)
        return rvx[layers, :], rvy[layers, :]

    def discharge(self, t, derivative=0):
        """The discharge in each layer.

        Parameters
        ----------
        t : scalar, list or array
            times at which discharge is computed.
            t must be ordered and tmin <= t <= tmax

        Returns
        -------
        array of discharges (nlayers,len(t))
            Discharge in each screen with zeros for layers that are not
            screened
        """
        # Could potentially be more efficient if s is pre-computed for
        # all elements, but may not be worthwhile to store as it is quick now
        time = np.atleast_1d(t).astype(float)
        rv = np.zeros((self.nlayers, len(time)))
        if self.type == "g":
            s = self.dischargeinflayers * self.model.p**derivative
            rv = invlapcomp(
                time,
                s[np.newaxis, :],
                self.model.npint,
                self.model.M,
                self.model.tintervals,
                np.zeros(self.ntstart, dtype="int"),
                self.tstart,
                self.bc,
                self.nlayers,
            )
        else:
            s = np.sum(self.parameters[:, :, np.newaxis, :] * self.dischargeinf, 1)
            s = np.sum(s[:, np.newaxis, :, :] * self.aq.eigvec, 2)
            s = s[:, self.layers, :] * self.model.p**derivative
            rv = invlapcomp(
                time,
                s,
                self.model.npint,
                self.model.M,
                self.model.tintervals,
                self.model.enumber,
                self.model.etstart,
                self.model.ebc,
                self.nlayers,
            )
        return rv

    # this function is kept for testing of version 0.6.6
    def dischargeold(self, t, derivative=0):
        """The discharge in each layer.

        Parameters
        ----------
        t : scalar, list or array
            times at which discharge is computed.
            t must be ordered and tmin <= t <= tmax

        Returns
        -------
        array of discharges (nlayers,len(t))
            Discharge in each screen with zeros for layers that are not
            screened
        """
        # Could potentially be more efficient if s is pre-computed for
        # all elements, but may not be worthwhile to store as it is quick now
        time = np.atleast_1d(t).astype(float)
        if (time[0] < self.model.tmin) or (time[-1] > self.model.tmax):
            print(
                "Warning, some of the times are smaller than tmin or"
                + "larger than tmax; zeros are substituted"
            )
        rv = np.zeros((self.nlayers, np.size(time)))
        if self.type == "g":
            s = self.dischargeinflayers * self.model.p**derivative
            for itime in range(self.ntstart):
                t = time - self.tstart[itime]
                for i in range(self.nlayers):
                    rv[i] += self.bc[itime] * self.model.inverseLapTran(s[i], t)
        else:
            s = np.sum(self.parameters[:, :, np.newaxis, :] * self.dischargeinf, 1)
            s = np.sum(s[:, np.newaxis, :, :] * self.aq.eigvec, 2)
            s = s[:, self.layers, :] * self.model.p**derivative
            for k in range(self.model.ngvbc):
                e = self.model.gvbclist[k]
                for itime in range(e.ntstart):
                    t = time - e.tstart[itime]
                    if t[-1] >= self.model.tmin:  # Otherwise all zero
                        for i in range(self.nlayers):
                            rv[i] += e.bc[itime] * self.model.inverseLapTran(s[k, i], t)
        return rv

    def headinside(self, t):
        print("This function not implemented for this element")
        return

    def storeinput(self, frame):
        self.inputargs, _, _, self.inputvalues = inspect.getargvalues(frame)

    def write(self):
        rv = self.name + "(" + self.model.modelname + ",\n"
        for key in self.inputargs[2:]:  # The first two are ignored
            if isinstance(self.inputvalues[key], np.ndarray):
                rv += (
                    key
                    + " = "
                    + np.array2string(self.inputvalues[key], separator=",")
                    + ",\n"
                )
            elif isinstance(self.inputvalues[key], str):
                rv += key + " = '" + self.inputvalues[key] + "',\n"
            else:
                rv += key + " = " + str(self.inputvalues[key]) + ",\n"
        rv += ")\n"
        return rv

    def run_after_solve(self):  # noqa: B027
        """Function to run after a solution is completed.

        For most elements nothing needs to be done, but for strings of elements some
        arrays may need to be filled.
        """

    @abstractmethod
    def plot(self, ax=None):
        """Plot the element."""



from typing import Optional

import matplotlib.pyplot as plt
import numpy as np

from ttim.aquifer import SimpleAquifer


class PlotTtim:
    def __init__(self, ml):
        self._ml = ml

    def topview(self, win=None, ax=None, figsize=None, layers=None):
        """Plot top-view.

        This method plots all elements (in specified layers).

        Parameters
        ----------
        win : list or tuple
            [x1, x2, y1, y2]
        ax : matplotlib.Axes, optional
            axes to plot on, default is None which creates a new figure
        figsize : tuple of 2 values
            size of figure
        layers : int or list of ints, optional
            layers to plot, default is None which plots elements in all layers

        Returns
        -------
        ax : matplotlib.Axes
            axes with plot
        """
        if ax is None:
            _, ax = plt.subplots(figsize=figsize)
            ax.set_aspect("equal", adjustable="box")
        if layers is not None and isinstance(layers, int):
            layers = [layers]
        for e in self._ml.elementlist:
            if layers is None or np.isin(e.layers, layers):
                e.plot(ax=ax)
        if win is not None:
            ax.axis(win)
        return ax

    def xsection(
        self,
        xy: Optional[list[tuple[float]]] = None,
        labels=True,
        params=False,
        ax=None,
        fmt=None,
    ):
        """Plot cross-section of model.

        Note: this method does not plot elements at this time. It does plot
        cross-section inhoms if the model is a cross-section model (ModelXsection).

        Parameters
        ----------
        xy : list of tuples, optional
            list of tuples with coordinates of the form [(x0, y0), (x1, y1)]. If not
            provided, a cross section with length 1 is plotted.
        labels : bool, optional
            add layer numbering labels to plot
        params : bool, optional
            add parameter values to plot
        ax : matplotlib.Axes, optional
            axes to plot on, default is None which creates a new figure
        fmt : str, optional
            format string for parameter values, e.g. '.2f' for 2 decimals.

        Returns
        -------
        ax : matplotlib.Axes
            axes with plot
        """
        if ax is None:
            _, ax = plt.subplots(1, 1, figsize=(8, 4))

        # if SimpleAquifer, plot inhoms and return
        if isinstance(self._ml.aq, SimpleAquifer):
            if xy is not None:
                (x1, _), (x2, _) = xy
            else:
                x1, x2 = ax.get_xlim()
            for inhom in self._ml.aq.inhomdict.values():
                inhom.plot(ax=ax, labels=labels, params=params, x1=x1, x2=x2, fmt=fmt)
            for e in self._ml.elementlist:
                e.plot(ax=ax)
            ax.set_xlim(x1, x2)
            ax.set_ylabel("elevation")
            ax.set_xlabel("x")
            return ax

        if fmt is None:
            fmt = ""
        ssfmt = ".2e"

        # else get cross-section line
        if xy is not None:
            (x0, y0), (x1, y1) = xy
            r = np.sqrt((x1 - x0) ** 2 + (y1 - y0) ** 2)
            if y0 == 0 and y1 == 0:
                r0 = x0
                ax.set_xlim(x0, x1)
            elif x0 == 0 and x1 == 0:
                ax.set_ylim(y0, y1)
                r0 = y0
            else:
                ax.set_xlim(0, r)
                r0 = 0.0
        else:
            r0 = 0.0
            r = 1.0

        # get values for layer and aquifer numbering
        if labels or params:
            lli = 1 if self._ml.aq.topboundary == "con" else 0
            aqi = 0
        else:
            lli = None
            aqi = None

        # plot layers
        for i in range(self._ml.aq.nlayers):
            # leaky layers
            if self._ml.aq.ltype[i] == "l":
                ax.axhspan(
                    ymin=self._ml.aq.z[i + 1],
                    ymax=self._ml.aq.z[i],
                    color=[0.8, 0.8, 0.8],
                )
                if labels:
                    ax.text(
                        r0 + 0.5 * r if not params else r0 + 0.25 * r,
                        np.mean(self._ml.aq.z[i : i + 2]),
                        f"leaky layer {lli}",
                        ha="center",
                        va="center",
                    )
                if params:
                    ax.text(
                        r0 + 0.75 * r if labels else r0 + 0.5 * r,
                        np.mean(self._ml.aq.z[i : i + 2]),
                        (
                            f"$c$ = {self._ml.aq.c[lli]:{fmt}}, "
                            f"$S_s$ = {self._ml.aq.Sll[lli]:{ssfmt}}"
                        ),
                        ha="center",
                        va="center",
                    )
                if labels or params:
                    lli += 1

            # aquifers
            if labels and self._ml.aq.ltype[i] == "a":
                ax.text(
                    r0 + 0.5 * r if not params else r0 + 0.25 * r,
                    np.mean(self._ml.aq.z[i : i + 2]),
                    f"aquifer {aqi}",
                    ha="center",
                    va="center",
                )
            if params and self._ml.aq.ltype[i] == "a":
                if aqi == 0 and self._ml.aq.phreatictop:
                    paramtxt = (
                        f"$k_h$ = {self._ml.aq.kaq[aqi]:{fmt}}, "
                        f"$S$ = {self._ml.aq.Saq[aqi]:{fmt}}"
                    )
                else:
                    paramtxt = (
                        f"$k_h$ = {self._ml.aq.kaq[aqi]:{fmt}}, "
                        f"$S_s$ = {self._ml.aq.Saq[aqi]:{ssfmt}}"
                    )
                if self._ml.aq.model3d:
                    paramtxt += f", $k_z/k_h$ = {self._ml.aq.kzoverkh[aqi]:{fmt}}"
                ax.text(
                    r0 + 0.75 * r if labels else r0 + 0.5 * r,
                    np.mean(self._ml.aq.z[i : i + 2]),
                    paramtxt,
                    ha="center",
                    va="center",
                )
            if (labels or params) and self._ml.aq.ltype[i] == "a":
                aqi += 1

        # aquifer-aquifer boundaries (for e.g. Model3D)
        for i in range(1, self._ml.aq.nlayers):
            if self._ml.aq.ltype[i] == "a" and self._ml.aq.ltype[i - 1] == "a":
                ax.axhspan(
                    ymin=self._ml.aq.z[i], ymax=self._ml.aq.z[i], color=[0.8, 0.8, 0.8]
                )
        # top and bottom
        ax.axhline(self._ml.aq.z[0], color="k", lw=0.75)
        ax.axhline(self._ml.aq.z[-1], color="k", lw=3.0)
        # add y-label
        ax.set_ylabel("elevation")
        # remove x-ticks if no coordinates provided
        if xy is None:
            ax.xaxis.set_ticks([])
            ax.xaxis.set_ticklabels([])
        return ax

    def head_along_line(
        self,
        x1=0,
        x2=1,
        y1=0,
        y2=0,
        npoints=100,
        t=1.0,
        layers=0,
        sstart=0,
        color=None,
        lw=1,
        figsize=None,
        ax=None,
        legend=True,
        grid=True,
    ):
        """Plot head along line.

        Parameters
        ----------
        x1, x2, y1, y2 : float
            start and end coordinates of line
        npoints : int
            number of points along line
        t : scalar or array
            times at which to plot heads
        layers :
            layers for which to plot heads
        sstart : float
            starting distance for cross-section
        color : str
            color of line
        lw : float
            line width
        figsize : tuple of 2 values
            size of figure
        ax : matplotlib.Axes
            axes to plot on, default is None which creates a new figure
        legend : bool
            add legend to plot
        grid : bool
            add grid to plot

        Returns
        -------
        ax : matplotlib.Axes
            axes with plot
        """
        layers = np.atleast_1d(layers)
        t = np.atleast_1d(t)
        if ax is None:
            _, ax = plt.subplots(figsize=figsize)
        x = np.linspace(x1, x2, npoints)
        y = np.linspace(y1, y2, npoints)
        s = np.sqrt((x - x[0]) ** 2 + (y - y[0]) ** 2) + sstart
        h = self._ml.headalongline(x, y, t, layers)
        nlayers, ntime, npoints = h.shape
        for i in range(nlayers):
            for j in range(ntime):
                lbl = f"head (t={t[j]}, layer={layers[i]})"
                if color is None:
                    ax.plot(s, h[i, j, :], lw=lw, label=lbl)
                else:
                    ax.plot(s, h[i, j, :], color, lw=lw, label=lbl)
        if legend:
            ax.legend(loc=(0, 1), ncol=3, frameon=False)
        if grid:
            ax.grid(True)
        return ax

    def contour(
        self,
        win,
        ngr=20,
        t=1,
        layers=0,
        levels=20,
        layout=True,
        labels=True,
        decimals=1,
        color=None,
        ax=None,
        figsize=None,
        legend=True,
    ):
        """Contour plot.

        Parameters
        ----------
        win : list or tuple
            [x1, x2, y1, y2]
        ngr : scalar, tuple or list
            if scalar: number of grid points in x and y direction
            if tuple or list: nx, ny, number of grid points in x and y direction
        t : scalar
            time
        layers : integer, list or array
            layers for which grid is returned
        levels : integer or array (default 20)
            levels that are contoured
        layout : boolean (default True_)
            plot layout of elements
        labels : boolean (default True)
            print labels along contours
        decimals : integer (default 1)
            number of decimals of labels along contours
        color : str or list of strings
            color of contour lines
        ax : matplotlib.Axes
            axes to plot on, default is None which creates a new figure
        figsize : tuple of 2 values (default is mpl default)
            size of figure
        legend : list or boolean (default True)
            add legend to figure
            if list of strings: use strings as names in legend

        Returns
        -------
        ax : matplotlib.Axes
            axes with plot
        """
        x1, x2, y1, y2 = win
        if np.isscalar(ngr):
            nx = ny = ngr
        else:
            nx, ny = ngr
        layers = np.atleast_1d(layers)
        xg = np.linspace(x1, x2, nx)
        yg = np.linspace(y1, y2, ny)
        h = self._ml.headgrid(xg, yg, t, layers)
        if ax is None:
            _, ax = plt.subplots(figsize=figsize)
            ax.set_aspect("equal", adjustable="box")
        # color
        if color is None:
            c = plt.rcParams["axes.prop_cycle"].by_key()["color"]
        elif isinstance(color, str):
            c = len(layers) * [color]
        elif isinstance(color, list):
            c = color
        if len(c) < len(layers):
            n = np.ceil(self._ml.aq.naq / len(c))
            c = n * c

        # contour
        cslist = []
        cshandlelist = []
        for i in range(len(layers)):
            cs = ax.contour(
                xg, yg, h[i, 0], levels, colors=c[i], negative_linestyles="solid"
            )
            cslist.append(cs)
            handles, _ = cs.legend_elements()
            cshandlelist.append(handles[0])
            if labels:
                fmt = f"%1.{decimals}f"
                ax.clabel(cs, fmt=fmt)
        if isinstance(legend, list):
            ax.legend(cshandlelist, legend, loc=(0, 1), ncol=3, frameon=False)
        elif legend:
            legendlist = ["layer " + str(i) for i in layers]
            ax.legend(cshandlelist, legendlist, loc=(0, 1), ncol=3, frameon=False)

        if layout:
            self.topview(win=[x1, x2, y1, y2], ax=ax)
        return ax



import matplotlib.pyplot as plt
import numpy as np

from ttim.element import Element
from ttim.equation import (
    FluxDiffEquation,
    HeadDiffEquation,
    HeadEquation,
    MscreenEquation,
)


class LineSink1DBase(Element):
    """LineSink1D Base Class.

    All LineSink1D elements are derived from this class

    Parameters
    ----------
    model : Model object
        Model to which the element is added
    xls : float
        x-coordinate of the line sink
    tsandbc : list of tuples
        list of tuples of the form (time, bc) for boundary conditions
    res : float
        resistance of the line sink
    wh : string or float
        wetted perimeter of the linesink, "H" for aquifer height,
        "2H" for 2x aquifer height (two-sided flow) or specify any float value
    layers : int, array or list
        layer (int) or layers (list or array) in which line sink is located
    type : string
        type of element, "g" for given, "v" for variable and  "z" for zero.
    name : string
        name of the element
    label : string, optional
        label of the element
    aq : Aquifer object
        aquifer in which the element is located
    inhomelement : boolean
        set to True if element is part of an inhomogeneity
    """

    tiny = 1e-8

    def __init__(
        self,
        model,
        xls=0,
        tsandbc=[(0, 1)],
        res=0,
        wh="H",
        layers=0,
        type="",
        name="LineSink1DBase",
        label=None,
        aq=None,
        inhomelement=False,
    ):
        super().__init__(
            model,
            nparam=1,
            nunknowns=0,
            layers=layers,
            tsandbc=tsandbc,
            type=type,
            name=name,
            label=label,
            inhomelement=inhomelement,
        )
        # Defined here and not in Element as other elements can have multiple
        # parameters per layers
        self.nparam = len(self.layers)
        self.xls = float(xls)
        self.res = np.atleast_1d(res).astype(np.float64)
        self.wh = wh
        self.aq = aq
        self.model.addelement(self)

    def __repr__(self):
        return self.name + " at " + str(self.xls)

    def initialize(self):
        # Control point to make sure the point is always the same for
        # all elements
        self.xc = np.array([self.xls])
        self.yc = np.zeros(1)
        self.ncp = 1
        if self.aq is None:
            self.aq = self.model.aq.find_aquifer_data(self.xc[0], self.yc[0])
        self.setbc()
        coef = self.aq.coef[self.layers, :]
        self.setflowcoef()
        # term is shape (self.nparam,self.aq.naq,self.model.npval)
        self.term = self.flowcoef * coef
        self.term2 = self.term.reshape(
            self.nparam, self.aq.naq, self.model.nint, self.model.npint
        )
        self.dischargeinf = self.flowcoef * coef
        self.dischargeinflayers = np.sum(
            self.dischargeinf * self.aq.eigvec[self.layers, :, :], 1
        )
        if isinstance(self.wh, str):
            if self.wh == "H":
                self.wh = self.aq.Haq[self.layers]
            elif self.wh == "2H":
                self.wh = 2.0 * self.aq.Haq[self.layers]
        else:
            self.wh = np.atleast_1d(self.wh) * np.ones(self.nlayers)
        # Q = (h - hls) / resfach
        self.resfach = self.res / (self.wh)
        # Q = (Phi - Phils) / resfacp
        self.resfacp = self.resfach * self.aq.T[self.layers]

    def setflowcoef(self):
        """Separate function so that this can be overloaded for other types."""
        self.flowcoef = 1.0 / self.model.p  # Step function

    def potinf(self, x, y=0, aq=None):
        """Can be called with only one x value."""
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((self.nparam, aq.naq, self.model.npval), dtype=complex)
        if aq == self.aq:
            if (x - self.xls) < 0.0:
                pot = -0.5 * aq.lab * np.exp((x - self.xls) / aq.lab)
            elif (x - self.xls) >= 0.0:
                pot = -0.5 * aq.lab * np.exp(-(x - self.xls) / aq.lab)
            else:
                raise ValueError("Something wrong with passed x value.")
            rv[:] = self.term * pot
        return rv

    def disvecinf(self, x, y=0, aq=None):
        """Can be called with only one x,y value."""
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        rvx = np.zeros((self.nparam, aq.naq, self.model.npval), dtype=complex)
        rvy = np.zeros((self.nparam, aq.naq, self.model.npval), dtype=complex)
        if aq == self.aq:
            if (x - self.xls) < 0.0:
                qx = 0.5 * np.exp((x - self.xls) / aq.lab)
            elif (x - self.xls) >= 0.0:
                qx = -0.5 * np.exp(-(x - self.xls) / aq.lab)
            else:
                raise ValueError("Something wrong with passed x value.")
            rvx[:] = self.term * qx
        return rvx, rvy

    def changetrace(
        self, xyzt1, xyzt2, aq, layer, ltype, modellayer, direction, hstepmax
    ):
        raise NotImplementedError("changetrace not implemented for this element")

    def plot(self, ax=None):
        if ax is None:
            _, ax = plt.subplots()
        aq = self.model.aq.find_aquifer_data(self.xls, 0.0)
        for ilay in self.layers:
            ax.plot(
                [self.xls, self.xls],
                [aq.zaqtop[ilay], aq.zaqbot[ilay]],
                "k-",
            )


class DischargeLineSink1D(LineSink1DBase):
    r"""Linesink1D with a specified discharge for each layer the linesink is in.

    Parameters
    ----------
    model : Model object
        model to which the element is added
    x : float
        x-coordinate of the linesink
    tsandq : list of tuples
        tuples of starting time and specific discharge after starting time
    res : float
        resistance of the linesink
    layers : int, array or list
        layer (int) or layers (list or array) in which linesink is located
    label : string or None (default: None)
        label of the linesink

    Examples
    --------
    Example of an infinitely long linesink that pumps with a specific discharge of
    100 between times 10 and 50, with a specific discharge of 20 between
    times 50 and 200, and zero speficic discharge after time 200.

    >>> DischargeLineSink1D(ml, tsandq=[(10, 100), (50, 20), (200, 0)])
    """

    def __init__(
        self, model, xls=0, tsandq=[(0, 1)], res=0, wh="H", layers=0, label=None
    ):
        super().__init__(
            model,
            xls,
            tsandbc=tsandq,
            res=res,
            wh=wh,
            layers=layers,
            type="g",
            name="DischargeLineSink1D",
            label=label,
        )


class LineSink1D(LineSink1DBase, MscreenEquation):
    r"""Linesink1D with a specified discharge.

    Parameters
    ----------
    model : Model object
        model to which the element is added
    x : float
        x-coordinate of the linesink
    tsandq : list of tuples
        tuples of starting time and specific discharge after starting time
    res : float
        resistance of the linesink
    layers : int, array or list
        layer (int) or layers (list or array) in which linesink is located
    label : string or None (default: None)
        label of the linesink

    Examples
    --------
    Example of an infinitely long linesink that pumps with a specific discharge of
    100 between times 10 and 50, with a specific discharge of 20 between
    times 50 and 200, and zero specific discharge after time 200.

    >>> LineSink1D(ml, tsandq=[(10, 100), (50, 20), (200, 0)])
    """

    def __init__(
        self,
        model,
        xls=0,
        tsandq=[(0, 1)],
        res=0,
        wh="H",
        vres=0.0,
        wv=1.0,
        layers=0,
        label=None,
    ):
        super().__init__(
            model,
            xls,
            tsandbc=tsandq,
            res=res,
            wh=wh,
            layers=layers,
            type="v",
            name="LineSink1D",
            label=label,
        )
        self.nunknowns = self.nparam
        # Vertical resistance inside line-sink
        self.vres = np.atleast_1d(vres)
        self.wv = wv
        if len(self.vres) == 1:
            self.vres = self.vres[0] * np.ones(self.nlayers - 1)

    def initialize(self):
        super().initialize()
        self.parameters = np.zeros(
            (self.model.ngvbc, self.nparam, self.model.npval), dtype=complex
        )
        # qv = (hn - hn-1) / vresfac[n - 1]
        self.vresfac = self.vres / self.wv


class HeadLineSink1D(LineSink1DBase, HeadEquation):
    """1D head-specified linesink element.

    Parameters
    ----------
    model : Model object
        Model to which the element is added
    xls : float
        x-coordinate of the linesink
    tsandh : list of tuples
        list of tuples of the form (time, head) for head conditions
    res : float
        resistance of the linesink
    wh : string or float
        wetted perimeter of the linesink, "H" for aquifer height,
        "2H" for 2x aquifer height (two-sided flow) or specify any float value
    layers : int, array or list
        layer (int) or layers (list or array) in which linesink is located
    label : string, optional
        label of the element
    """

    def __init__(
        self,
        model,
        xls=0,
        tsandh=[(0, 1)],
        res=0,
        wh="H",
        layers=0,
        label=None,
    ):
        super().__init__(
            model,
            xls,
            tsandbc=tsandh,
            res=res,
            wh=wh,
            layers=layers,
            type="v",
            name="HeadLineSink1D",
            label=label,
        )
        self.nunknowns = self.nparam

    def initialize(self):
        super().initialize()
        self.parameters = np.zeros(
            (self.model.ngvbc, self.nparam, self.model.npval), dtype=complex
        )
        # Needed in solving, solve for a unit head
        self.pc = self.aq.T[self.layers]


class HeadDiffLineSink1D(LineSink1DBase, HeadDiffEquation):
    """1D head-difference linesink element.

    Used to ensure continuity of head in a cross-section model, e.g. at the boundary
    of an inhomogeneity.

    Parameters
    ----------
    model : Model object
        Model to which the element is added
    xls : float
        x-coordinate of the linesink
    layers : int, array or list
        layer (int) or layers (list or array) in which linesink is located
    label : string, optional
        label of the element
    aq : Aquifer object
        aquifer in which the element is located
    """

    def __init__(
        self,
        model,
        xls=0,
        layers=0,
        label=None,
        aq=None,
    ):
        super().__init__(
            model,
            xls,
            tsandbc=[(0, 0)],
            res=0.0,
            wh="H",
            layers=layers,
            type="z",
            name="HeadDiffLineSink1D",
            label=label,
            aq=aq,
            inhomelement=True,
        )
        self.nunknowns = self.nparam

    def initialize(self):
        super().initialize()
        self.xcout = self.xc + self.tiny
        self.xcin = self.xc - self.tiny
        self.ycout = np.zeros(1)
        self.ycin = np.zeros(1)
        self.cosout = -np.ones(1)
        self.sinout = np.zeros(1)
        self.aqout = self.model.aq.find_aquifer_data(self.xcout[0], 0)
        self.aqin = self.model.aq.find_aquifer_data(self.xcin[0], 0)

        self.parameters = np.zeros(
            (self.model.ngvbc, self.nparam, self.model.npval), dtype=complex
        )

    def plot(self, ax=None):
        if ax is None:
            _, ax = plt.subplots()
        aqout = self.model.aq.find_aquifer_data(self.xcout, 0.0)
        aqin = self.model.aq.find_aquifer_data(self.xcin, 0.0)
        ztop_out = aqout.z[0]
        ztop_in = aqin.z[0]

        ax.plot(
            [self.xls, self.xls],
            [np.max([ztop_in, ztop_out]), aqout.z[-1]],
            "k--",
            lw=1.0,
        )
        return ax


class FluxDiffLineSink1D(LineSink1DBase, FluxDiffEquation):
    """1D flux-difference linesink element.

    Used to ensure continuity of flux in a cross-section model, e.g. at the boundary
    of an inhomogeneity.

    Parameters
    ----------
    model : Model object
        Model to which the element is added
    xls : float
        x-coordinate of the linesink
    layers : int, array or list
        layer (int) or layers (list or array) in which linesink is located
    label : string, optional
        label of the element
    """

    def __init__(self, model, xls=0, layers=0, label=None, aq=None):
        super().__init__(
            model,
            xls,
            tsandbc=[(0, 0)],
            res=0.0,
            wh="H",
            layers=layers,
            type="z",
            name="FluxDiffLineSink1D",
            label=label,
            aq=aq,
            inhomelement=True,
        )
        self.nunknowns = self.nparam

    def initialize(self):
        super().initialize()
        self.xcout = self.xc - self.tiny
        self.xcin = self.xc + self.tiny
        self.ycout = np.zeros(1)
        self.ycin = np.zeros(1)
        self.cosout = -np.ones(1)
        self.sinout = np.zeros(1)
        self.aqout = self.model.aq.find_aquifer_data(self.xcout[0], 0)
        self.aqin = self.model.aq.find_aquifer_data(self.xcin[0], 0)
        self.parameters = np.zeros(
            (self.model.ngvbc, self.nparam, self.model.npval), dtype=complex
        )

    def plot(self, ax=None):
        if ax is None:
            _, ax = plt.subplots()
        aqout = self.model.aq.find_aquifer_data(self.xcout, 0.0)
        aqin = self.model.aq.find_aquifer_data(self.xcin, 0.0)
        ztop_out = aqout.z[0]
        ztop_in = aqin.z[0]

        ax.plot(
            [self.xls, self.xls],
            [np.max([ztop_in, ztop_out]), aqout.z[-1]],
            "k--",
            lw=1.0,
        )
        return ax
